;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

(in-package :stella)

;;;
;;; font utilities

(defvar %soi-height% nil)
(defvar %soi-spacing% nil)
 

(defstruct (%font (:type (vector fixnum))) 
  (ff 0)
  (ms 0)
  (ascent 0)
  (descent 0)
  (maximum-width 0)
  (leading 0))

(defvar %capella-font% (make-%font))

(defun %compile-font-spec (font-spec %font-struct)
  (multiple-value-bind (ff ms) (ccl:font-codes font-spec)
    (ccl:rlet ((fmi :fminput
                :family (ldb (byte 16 16) ff)
                :size (ldb (byte 16 0) ms)
                :face (ldb (byte 16 0) ff)
                :needbits nil
                :device 0
                :numer #x10001
                :denom #x10001))
      (let ((fmo (#_fmswapfont fmi)))
        
        (loop for i below 6 
              for val in (list ff 
                               ms 
                               (ccl:rref fmo fmoutput.ascent)
                               (ccl:rref fmo fmoutput.descent)
                               (ccl:rref fmo fmoutput.widmax)
                               (ccl:rref fmo fmoutput.leading))
              do (setf (aref %font-struct i) val))
        )))
  font-spec)

(defun set-capella-font (font-spec)
  (%compile-font-spec font-spec %capella-font%)
  (setf *capella-font* font-spec)
  (setf %soi-height% (+ (%font-ascent %capella-font%)
                        (%font-descent %capella-font%)
                        (%font-leading %capella-font%)
                        12)
        %soi-spacing% (+ %soi-height% 2))
  *capella-font*)

(eval-when (:load-toplevel :execute)
  ;(set-capella-font '("Geneva" 10 :srcor :plain))
  (set-capella-font *capella-font*)
  )


(defmacro draw-using-font (font &body body)
  `(multiple-value-bind (ff ms) (ccl:grafport-font-codes)
     (unwind-protect 
       (progn
         (ccl:set-grafport-font-codes (%font-ff ,font) (%font-ms ,font))
         ,@body)
       (ccl:set-grafport-font-codes ff ms))))


;;;
;;; color utilities
;;;

;;;
;;; a highlighted color is a 25% reduction in brightness.  colors which
;;; have less than 25% brightness, however, wrap around and are mapped
;;; onto the range between 100% and 75% by setting the maximum component
;;; to its complement and scaling the remaining components accordingly.
;;; white becomes black and black white.
;;;
;;; although this is not the most elegant approach, it seems to give
;;; satisfactory results.

(defun hilited-color (color)
  (if (= color 0)
    #xffffff
    (if (= color #xffffff)
      0
      (let* ((new-color #x000000)
             (r (ldb #xff0000 color))
             (g (ldb #xff00 color))
             (b (ldb #xff color))
             (max (max r g b)))
        (if (>= max #x40)
          (loop for bsp in '(#xff0000 #xff00 #xff)
                for c in (list r g b)
                do (setf new-color (dpb (- c (ash c -2))
                                        bsp new-color)))
          (if (> max 0)
            (loop for bsp in '(#xff0000 #xff00 #xff)
                  for c in (list r g b)
                  with comp = (- #xff max)
                  do (setf new-color (dpb (round (* (/ c max) comp))
                                          bsp new-color)))
            (setf new-color #xffffff)))
        ;(format t "~&converted color #x~6,'0x to #x~6,'0x" color new-color)
        new-color))))

#|
(defun hilited-color (color)
  (let ((new-color #x000000))
    (loop for bsp in '(#xff #xff00 #xff0000)
          for b = (ldb bsp color)
          do (setf new-color (dpb (- b #x40) bsp new-color)))
    (format t "~&converted color #x~6,'0x to #x~6,'0x" color new-color)
    new-color))
|#

;(format t "~&#x~6,'0x" ccl:*red-color*)
;(hilited-color ccl:*tan-color*)

;;;
;;; a disabled color implies a 50% percent decrease in saturation.
;;; we do that by halving the distance of the brightest component to 
;;; the other two.
;;; i realize this doesn't have any effect on grayscale colors, but i 
;;; can't think about a better way...
;;; using ccl:*gray-pattern* as a pen pattern while drawing with 
;;; disabled colors is therfore highlly recommended.

(defun disabled-color (color)
  (let* ((new-color #x000000)
         (r (ldb #xff0000 color))
         (g (ldb #xff00 color))
         (b (ldb #xff color))
         (max (max r g b)))
    (loop for bsp in '(#xff0000 #xff00 #xff)
          for c in (list r g b)
          do (setf new-color (dpb (+ c (ash (- max c) -1)) bsp new-color)))
    ;(format t "~&converted color #x~6,'0x to #x~6,'0x" color new-color)
    new-color))

;(format t "~&#x~6,'0x" ccl:*red-color*)
;(disabled-color ccl:*red-color*)

  
;;;
;;; this should be used to do all the color jobs inside hilite-view-methods

(defmacro with-appropriate-fore-color (color hilite enabled &body body)
  `(ccl:with-fore-color (if ,enabled 
                          (if ,hilite
                            (hilited-color ,color)
                            ,color)
                          (disabled-color ,color))
     ,@body))




;;;
;;;
;;; controls
;;;
;;; 


;;;
;;;   Classes:
;;;   ^^^^^^^^                                 [ccl:simple-view]
;;;                                             |
;;;                                             |
;;;              titled-item-mixin:             item:
;;;                - title                        - framed?
;;;              :                                - frame-curvature
;;;              :                                - background-color
;;;              :                              |
;;;     _________:______________________________|
;;;    |         :                              |
;;;    |         :                              control:
;;;    |         :                                - action
;;;    |         :                                - enabled?
;;;    |         :                              |
;;;   ...........:..............................| 
;;;   :|         :                             :|
;;;   :|         : ____________________________:|_______________
;;;   :|         :|                            :|               |
;;;   label:     toggle:                       button:          icon-button:
;;;                - selected?                   - default?       - icon-id
;;;              |                             |
;;;         _____|_______                      |
;;;         |            |                     |
;;;         check-box:   radio-button:         default-button:
;;;                        - cluster           |
;;;                                            |
;;;                                            invisible-default-button:
;;;



;;;
;;; general methods
;;; 

(defmethod control-enable ((self ccl:simple-view)) self)
(defmethod control-disable ((self ccl:simple-view)) self)
(defmethod control-enabled? ((self ccl:simple-view)) self)


(defmethod control-disable ((self ccl:dialog-item))
  (ccl:dialog-item-disable self))

(defmethod control-enable ((self ccl:dialog-item))
  (ccl:dialog-item-enable self))

(defmethod control-enabled? ((self ccl:dialog-item))
  (ccl:dialog-item-enabled-p self))


(defmethod item-title ((self ccl:dialog-item))
  (ccl:dialog-item-text self))

(defmethod set-item-title ((self ccl:dialog-item) title)
  (ccl:set-dialog-item-text self title)
  (ccl:invalidate-view self))


;;;
;;; items are generic user interface items with an optional
;;; frame (a rounded rect).
;;; the value of :framed can be nil or any integer >= 0.
;;; nil denotes no frame, 0 a rectangular one, and any other
;;; the diameter of the corner curves.

(defclass item (ccl:simple-view)
  ((framed? :initarg :framed? :initform t :accessor item-framed?)
   (frame-curvature :initarg :frame-curvature :initform 0 
              :accessor item-frame-curvature)
   (background-color :initarg :background-color :initform nil 
               :accessor item-background-color))
  (:default-initargs
    :view-font *capella-font*))

(defmethod control-enable ((self item)) self)
(defmethod control-disable ((self item)) self)
(defmethod control-enabled? ((self item)) self)

(defmethod item-title ((self item))
  (declare (ignore self)))

(defmethod set-item-title ((self item) title)
  (declare (ignore self title)))

(defun item-installed? (item)
  (let ((dialog (ccl:view-container item)))
    (and dialog (ccl:wptr dialog))))

;;;
;;; generic control 

(defclass control (item)
  ((action :initarg :action :initform nil :accessor control-action)
   (enabled? :initarg :enabled? :initform t :accessor control-enabled?))
  (:default-initargs
    :view-font *capella-font*))

(defmethod control-enable ((view control))
  (%enable-control view t))

(defmethod control-disable ((view control))
  (%enable-control view nil))

(defmethod %enable-control ((view control) enabled?)
  (unless (eq enabled? (control-enabled? view))
    (setf (control-enabled? view) enabled?)
    (ccl:invalidate-view view)))

(defmethod ccl:view-draw-contents ((self control)) 
  (call-next-method)
  (hilite-view self nil (control-enabled? self)))

(defmethod update-control ((self control))
  (declare (ignore self))
  )

(defmethod ccl:view-click-event-handler ((self control) where)
  (declare (ignore where))
  (when (control-enabled? self)
    (let ((con (ccl:view-container self))
          (in-rect t))
      (when (loop initially (hilite-view self t t)
                  do 
                  (when (not (ccl:mouse-down-p))
                    (return in-rect))
                  (if (ccl:view-contains-point-p 
                       self (ccl:view-mouse-position con))
                    (when (not in-rect)
                      (setf in-rect t)
                      (hilite-view self t t))
                    (when in-rect
                      (setf in-rect nil)
                      (hilite-view self nil t)))
                  finally (return in-rect))
        (update-control self)
        (unwind-protect
          (funcall (or (control-action self) #'identity) self)
          (hilite-view self nil t))
        t))))

(defmethod worksheet-action ((view control))
  (worksheet-action (view-worksheet view)))

(defmethod update-worksheet ((view control))
  (update-worksheet (view-worksheet view)))


;;;
;;; hilite-view does the actual work of drawing the control

(defgeneric hilite-view (self &optional hilite enabled))

(defmethod hilite-view ((self control) &optional hilite enabled)
  (ccl:with-focused-view self
    (let ((s (ccl:view-size self)))
      (ccl:rlet ((r :rect :topleft 0 :bottomright s)
                 (ps :penstate))
        (#_getpenstate ps)
        (#_pennormal)
        (#_penpat (if enabled ccl:*black-pattern* ccl:*gray-pattern*))
        (let ((c (or (item-frame-curvature self) 0))
              (bc (item-background-color self)))
          (when bc
            (with-appropriate-fore-color bc hilite enabled 
              (#_fillroundrect r c c ccl:*black-pattern*)))
          (when (item-framed? self)
            (with-appropriate-fore-color ccl:*black-color* hilite enabled 
              (#_frameroundrect r c c))))
        (#_setpenstate ps)))))


;;;
;;; abstract class

;;;
;;; pane is the root class for items containing other items

(defclass pane (item ccl:view)
  ())

(defclass titled-item-mixin ()
  ((title :initarg :title :initform nil :accessor item-title)
   (title-justification :initarg :title-justification :initform :left 
                        :accessor titled-item-title-justification))
  (:default-initargs :title ""))


;; why do i need this one when its an slot accessor?
(defmethod item-title ((self titled-item-mixin))
  (slot-value self 'title))

(defmethod set-item-title ((self titled-item-mixin) title)
  (setf (item-title self) title)
  (ccl:invalidate-view self))

;;;
;;; labels are static text elements

(defclass label (titled-item-mixin item)
  ()
  (:default-initargs
    :framed? nil
    :view-font *capella-font*))

(defmethod ccl:view-default-size ((self label))
  (ccl:with-pstrs ((s (or (item-title self) "Untitled")))
    (draw-using-font %capella-font%
        (let ((sw (#_stringwidth s))
              (sh (+ (%font-ascent %capella-font%)
                     (%font-descent %capella-font%))))
          (xy sw sh)))))

(defmethod ccl:view-draw-contents ((self label))
  (ccl:with-focused-view self
    (let ((s (ccl:view-size self))
          (f (item-framed? self)))
      (ccl:rlet ((or :rect :topleft 0 :bottomright s)
                 (ir :rect :topleft 0 :bottomright s)
                 (ps :penstate))
        (#_getpenstate ps)
        (#_pennormal)
        (when f (#_insetrect ir 2 2))
        (let ((c (or (item-frame-curvature self) 0))
              (bc (item-background-color self)))
          (when bc
            (ccl:with-fore-color bc
              (#_fillroundrect or c c ccl:*black-pattern*)))
          (ccl:with-fore-color ccl:*black-color*
            (when f (#_frameroundrect or c c))
            ;; draw the title
            (when (item-title self)
              (ccl:with-pstrs ((str (item-title self)))
                ;; draw the centered title:
                (draw-using-font %capella-font%
                  (let* ((w (xy-x s))
                         (h (xy-y s))
                         (tw (#_stringwidth str))
                         (ta (%font-ascent %capella-font%))
                         (corr (if f 2 0)))
                    (#_moveto (ecase (titled-item-title-justification self)
                                (:left corr)
                                (:center (- (floor w 2) (floor tw 2)))
                                (:right (- w (+ tw corr))))
                     (+ (ceiling (- h ta (%font-descent %capella-font%)) 2) 
                        ta))
                    (ccl:with-clip-rect ir
                      (#_drawstring str))))))))
        (#_setpenstate ps)))))


;;;
;;; buttons are controls with a title and a "standard" curvature
;;; default-buttons have an outline and a default title "OK"

(defclass button (titled-item-mixin control)
  ((default? :initarg :default? :initform nil :accessor button-default?))
  (:default-initargs
    :frame-curvature 10
    :title-justification :center
    :background-color ccl:*white-color*))

(defclass default-button (button)
  ()
  (:default-initargs 
    :default? t
    :title "OK"
    :view-nick-name ':Ok))

(defmethod ccl:view-default-size ((self button))
  (ccl:with-pstrs ((s (item-title self)))
    (draw-using-font %capella-font%
      (let ((sw (#_stringwidth s))
            (sh (+ (%font-ascent %capella-font%)
                   (%font-descent %capella-font%)))
            (l (%font-leading %capella-font%))
            (w (%font-maximum-width %capella-font%))
            (def (if (button-default? self) 6 0)))
        (xy (+ w sw w def) (+ l sh l def))))))

(defmethod ccl:dialog-item-enabled-p ((self default-button))
  (control-enabled? self))

(defmethod ccl:press-button ((self default-button))
  (control-push self))

(defmethod hilite-view ((self button) &optional hilite enabled)
  ;(ccl:set-view-font self *button-font*)
  (ccl:with-focused-view self 
    (let ((s (ccl:view-size self))
          (default? (button-default? self)))
      (ccl:rlet ((or :rect :topleft 0 :bottomright s)
                 (ir :rect :topleft 0 :bottomright s)
                 (ps :penstate))
        ;; default buttons have a different inner rectangle
        (when default? (#_insetrect ir 3 3))
        (#_getpenstate ps)
        (#_pennormal)
        ;; c is inner curvature, oc outer
        (let* ((c (or (item-frame-curvature self) 0))
               (oc (if default? 
                     (if (= c 0) 0 (+ c 6))
                     c))
               (bc (item-background-color self)))
          
          ;; background: use inner rectangle for default buttons
          (when bc
            (with-appropriate-fore-color bc hilite enabled
              (#_fillroundrect ir c c ccl:*black-pattern*)))
          
          ;; default buttons draw an outline using the outer rectangle.
          ;; Note that the outline gets NEVER highlighted.
          (when default? 
            (with-appropriate-fore-color ccl:*black-color* nil enabled
              (#_pensize 2 2)
              (#_frameroundrect or oc oc)
              (#_pensize 1 1)))
          
          (with-appropriate-fore-color ccl:*black-color* hilite enabled 
            ;; frame the control using inner rectangle for default buttons
            (when (item-framed? self)
              (#_frameroundrect ir c c))
            (when (item-title self)
              (ccl:with-pstrs ((str (item-title self)))
                ;; draw the centered title:
                ;; ascent+descent is a whole line height, INCLUDING leading.
                ;; assume the ACTUAL used ascent to be only about 85% of
                ;; the ascent reported by ccl:font-info
                (draw-using-font %capella-font%
                  (let* ((w (xy-x s))
                         (h (xy-y s))
                         (tw (#_stringwidth str))
                         (ta (%font-ascent %capella-font%))
                         ;; ? im not sure about the 6
                         (corr (if default? 6 4)))      
                    ;; default buttons have a smaller usable height
                    ;(when default? (decf h 6)) -- fucks up
                    (#_moveto (ecase (titled-item-title-justification self)
                                (:left corr)
                                (:center (- (floor w 2) (floor tw 2)))
                                (:right (- w (+ tw corr))))
                     (+ (ceiling (- h ta (%font-descent %capella-font%)) 2) ta))
                    (ccl:with-clip-rect ir
                      (#_drawstring str))))))
            ;; dim the control using the outer rectangle
            (unless enabled
              (#_penpat ccl:*gray-pattern*) 
              ;; :patBic (11) or :notpatBic (15), but keywords don't work :(
              (#_penmode 15)
              (#_paintroundrect or oc oc))
            )
          
          ;; hilite the inner rectangle, if necessary
          ;(when hilite (#_invertroundrect ir c c))
          )
        (#_setpenstate ps)))))

;;;
;;; make return and enter keys trigger our action
;;; i don't check for other default-buttons in the window (mcl doesn't 
;;; bother either)

(defmethod ccl:install-view-in-window :before ((self default-button) view)
  (declare (ignore view))
  (let ((dialog (ccl:view-window self)))
    (when dialog
      (if (null (ccl::%get-default-button dialog))
        (setf (ccl::%get-default-button dialog) self)))))

(defmethod ccl:remove-view-from-window :before ((self default-button))
  (let ((dialog (ccl:view-window self)))
    (when dialog
      (when (eq self (ccl::%get-default-button dialog))
        (setf (ccl::%get-default-button dialog) nil)))))

(defmethod control-push ((self control))
  (ccl:with-focused-dialog-item (self)
    (hilite-view self t t)
    (let ((time (+ 5 (#_tickcount))))
      (loop while (< (#_tickcount) time)))
    (hilite-view self nil t)
    (funcall (control-action self) self)))

;;;
;;; invisible default-buttons just trigger an action

(defclass invisible-default-button (default-button)
  ()
  (:default-initargs 
    :default? t
    :view-size 0
    :view-position 0
    :title ""))

(defmethod ccl:view-default-size ((self invisible-default-button))
  (xy 0 0))

(defmethod hilite-view ((self invisible-default-button) 
                        &optional hilite enabled)
  (declare (ignore self hilite enabled))
  )


;;;
;;; icon-buttons

(defclass icon-button (control)
  ((icon-id :initarg :icon-id :accessor icon-button-icon-id 
            :initform 128)))            ; displays the ccl icon


(defmethod hilite-view ((self icon-button) &optional hilite enabled)
  (let* ((s (ccl:view-size self))
         (w (1- (xy-x s)))
         (h (1- (xy-y s))))
    (ccl:with-focused-view self
      (ccl:rlet ((r :rect :topleft 0 :bottomright s)
                 (ps :penstate))
        (#_getpenstate ps)
        (#_insetrect r 1 1)
        (#_PlotIconID r 0        ; = $atNone = no alignment
         (logior (if hilite #x4000 0) (if enabled 0 1))
         (icon-button-icon-id self))
        (ccl:with-fore-color (if enabled
                               (if hilite ccl:*black-color*
                                   ccl:*white-color*)
                               ccl:*gray-color*)
          (#_moveto 0 h)
          (#_lineto 0 0)
          (#_lineto w 0))
        (ccl:with-fore-color (if enabled
                               (if hilite ccl:*white-color* 
                                   ccl:*black-color*)
                               ccl:*gray-color*)
          (#_moveto w 1)
          (#_lineto w h)
          (#_lineto 0 h))
        (#_setpenstate ps)))))



;;;
;;;  toggle is an abstract class

(defclass toggle (titled-item-mixin control)
  ((selected? :initarg :selected? :initform nil :accessor toggle-selected?)))


;;;
;;; standard toggles

(defun check-box-and-radio-button-default-size (self)
  (ccl:with-pstrs ((s (item-title self)))
    (draw-using-font %capella-font%
      (let ((sw (#_stringwidth s))
            (sh (+ (%font-ascent %capella-font%)
                   (%font-descent %capella-font%)))
            (w (%font-maximum-width %capella-font%)))
        (xy (+ 12 (ceiling w 2) sw) (max 12 sh))))))


;;;
;;; checkboxes have a 12 pixel square area at the left and a title

(defclass check-box (toggle)
  ())

(defmethod ccl:view-default-size ((self check-box))
  (check-box-and-radio-button-default-size self))

(defmethod toggle-select ((self check-box))
  (setf (toggle-selected? self) t)
  (when (item-installed? self)
    (hilite-view self nil (control-enabled? self))))

(defmethod toggle-deselect ((self check-box))
  (setf (toggle-selected? self) nil)
  (when (item-installed? self)
    (hilite-view self nil (control-enabled? self))))

(defmethod update-control ((self check-box))
  (setf (toggle-selected? self) (not (toggle-selected? self))))

(defmethod hilite-view ((self check-box) &optional hilite enabled)
  (ccl:without-interrupts 
    (let* ((s (ccl:view-size self))       ; size
           (h (xy-y s))                   ; height
           (bt (round (/ (- h 12) 2)))  ; box-top
           (bc (item-background-color self)))
      (ccl:with-focused-view self
        (ccl:rlet ((r :rect               ; control rectangle
                      :topleft 0 :bottomright s)
                   (b :rect               ; box
                      :left 0 :top bt :bottomright (xy 12 (+ 12 bt)))
                   (ps :penstate))
          (#_getpenstate ps)
          (#_pennormal)
          
          ;; fill background
          (when bc
            (ccl:with-fore-color bc
              (#_fillrect r ccl:*black-pattern*)))
          
          ;; fill the box white
          (ccl:with-fore-color ccl:*white-color*
            (#_fillrect b ccl:*black-pattern*))
          
          (ccl:with-fore-color ccl:*black-color*
            ;; draw the checkbox frame -- draw a thick outline if hilighted
            (if hilite 
              (progn 
                (#_pensize 2 2)
                (#_framerect b)
                (#_pensize 1 1))
              (#_framerect b))
            ;; draw selected? status
            (when (toggle-selected? self)
              (#_moveto 0 bt)
              (#_line 11 11)
              (#_moveto 11 bt)
              (#_line -11 11))
            (when (item-title self)
              (draw-using-font %capella-font%
                (ccl:with-pstrs ((str (item-title self)))
                  ;; draw title left justified/vertically centered
                  (let ((a (%font-ascent %capella-font%)))
                    (#_moveto 
                     (+ 12 (floor (%font-maximum-width %capella-font%) 2)) 
                     (+ (ceiling (- h a (%font-descent %capella-font%)) 2) a)))
                  (#_drawstring str))))
            ;; dim the control using the outer rectangle
            (unless enabled
              (#_penpat ccl:*gray-pattern*) 
              ;; :patBic (11) or :notpatBic (15), but keywords don't work :(
              (#_penmode 15)
              (#_paintrect r)))
          
          (#_setpenstate ps))))))



;;;
;;; radio-buttons have a round control instead of a square and
;;; a cluster id

(defclass radio-button (toggle)
  ((cluster :initarg :cluster :initform 0 :accessor radio-button-cluster)))

(defmethod ccl:view-default-size ((self radio-button))
  (check-box-and-radio-button-default-size self))

(defmethod pushed-radio-button ((view ccl:view) &optional (cluster 0))
  (ccl:dovector (item (ccl:view-subviews view))
    (when (and (typep item 'radio-button)
               (eq cluster (radio-button-cluster item))
               (toggle-selected? item))
      (return-from pushed-radio-button item))))

(defmethod toggle-select ((me radio-button))
  (unless (toggle-selected? me)
    (when (item-installed? me)
      (let ((cont (ccl:view-container me))
            (id (radio-button-cluster me)))
        (setf (toggle-selected? me) t)
        (ccl:map-subviews 
         cont #'(lambda (rb) 
                  (when (eql (radio-button-cluster rb) id)
                    (when (toggle-selected? rb)
                      (unless (eq rb me)
                        (setf (toggle-selected? rb) nil)
                        (hilite-view rb nil (control-enabled? rb))))) )
         'radio-button)))))

(defmethod toggle-deselect ((item radio-button))
  (setf (toggle-selected? item) nil)
  (when (item-installed? item)
    (hilite-view item nil (control-enabled? item))))

;(defun %debug-radio (rb &optional (str ""))
;  (format t "~&~a~a is ~:[off~;on~] and ~:[dis~;en~]abled." 
;          str (item-title rb) 
;          (toggle-selected? rb) (control-enabled? rb)))

;(defmethod ccl:view-click-event-handler :around ((self radio-button) where)
;  (declare (ignore where)) 
;  (when (call-next-method)
;    ;(%debug-radio self "Entry:  ")
;    (toggle-select self)))

(defmethod update-control ((self radio-button))
  (toggle-select self))


(defmethod hilite-view ((self radio-button) &optional hilite enabled)
  (ccl:without-interrupts 
    (let* ((s (ccl:view-size self))       ; size
           (h (xy-y s))                   ; height
           (rt (round (/ (- h 12) 2)))    ; circle-top
           (bc (item-background-color self)))
      (ccl:with-focused-view self
        (ccl:rlet ((r :rect               ; control rectangle
                      :topleft 0 :bottomright s)
                   (c :rect               ; circle
                      :left 0 :top rt :bottomright (xy 12 (+ 12 rt)))
                   (ps :penstate))
          (#_getpenstate ps)
          (#_pennormal)
          
          ;; fill background
          (when bc
            (ccl:with-fore-color bc
              (#_fillrect r ccl:*black-pattern*)))
          
          ;; fill the circle white
          (ccl:with-fore-color ccl:*white-color*
            (#_fillroundrect c 12 12 ccl:*black-pattern*))
          
          (ccl:with-fore-color ccl:*black-color*
            ;; draw the circle frame -- draw a thicker outline if hilighted
            (if hilite 
              (progn 
                (#_pensize 2 2)
                (#_frameroundrect c 12 12)
                (#_pensize 1 1))
              (#_frameroundrect c 12 12))
            ;; draw status (selected?)
            (when (toggle-selected? self)
              (#_insetrect c 3 3)
              (#_fillroundrect c 6 6 ccl:*black-pattern*))
            (when (item-title self)
              (draw-using-font %capella-font%
                (ccl:with-pstrs ((str (item-title self)))
                  ;; draw title left justified/vertically centered
                  (let ((a (%font-ascent %capella-font%)))
                    (#_moveto 
                     (+ 12 (floor (%font-maximum-width %capella-font%) 2)) 
                     (+ (ceiling (- h a (%font-descent %capella-font%)) 2) a)))
                  (#_drawstring str))))
            ;; dim the control using the outer rectangle
            (unless enabled
              (#_penpat ccl:*gray-pattern*) 
              ;; :patBic (11) or :notpatBic (15), but keywords don't work :(
              (#_penmode 15)
              (#_paintrect r)))
          
          (#_setpenstate ps))))))



;;;
;;; these are still based on ccl items
;;; 


(defclass textbuf (ccl:editable-text-dialog-item)
  ((ccl:dialog-item-text :initarg :title)
   (ccl:dialog-item-action-function :initarg :action)
   (ccl::draw-outline :initarg :framed?))
  (:default-initargs 
    :view-font *capella-font*))

(defun textbuf-text (buf)
  (string-trim '(#\Space #\Tab #\Return) (ccl:dialog-item-text buf)))

(defun textbuf-integer (textbuf &optional default)
  (let ((text (textbuf-text textbuf)))
    (multiple-value-bind (int pos)
                         (parse-integer text :junk-allowed t)
      (if (= pos (length text)) 
        (values (or int default) t)
        (values nil nil)))))

(defun textbuf-float (buf &optional default)
  (multiple-value-bind (value flag) (textbuf-form buf default)
    (if (numberp value)
      (values (coerce value 'float) flag)
      (values value flag))))

(defun textbuf-form (buf &optional default)
  (let ((text (textbuf-text buf)))
    (if (string= text "")
      (values default t)
      (if (balanced-form? text)
        (values (read-from-string text) t)
        (values nil nil)))))



(defun textbuf-function (buf &optional default)
  (let ((thing (textbuf-form buf default)))
    (cond ((null thing) 
           (values nil nil))
          ((symbolp thing)
           (if (fboundp thing)
             (values (symbol-function thing) t)
             (values nil nil)))
          ((consp thing)
           (if (eq (first thing) 'lambda)
             (values (coerce thing 'function) t)
             (if (eq (first thing) 'function)
               (values (coerce (second thing) 'function) t)
               (values nil nil))))
          ((typep thing 'function)
           (values thing t))
          (t (values nil nil)))))

(defclass pop-up-menu (ccl:pop-up-menu)
  ((ccl:dialog-item-enabled-p :initarg :enabled?))
  (:default-initargs :view-font *capella-font*))


;;;
;;;
;;;

(defclass output-stream-edit-button (button)
  ((stream :initform nil :initarg :stream 
           :accessor output-stream-edit-button-stream))
  (:default-initargs
    :title +slot-unset+
    :title-justification :left
    :frame-curvature 0
    :view-nick-name ':output-stream-button
    :background-color *bgc*
    :action #'(lambda (self)
                (let ((stream (output-stream-edit-button-stream self)))
                  (and stream (browser stream 'info-browser))))))

(defmethod current-output-stream ((b output-stream-edit-button))
  (output-stream-edit-button-stream b))
    
(defun output-stream-edit-button-select (button enter?)
  (%draw-mouse-sensitive-rect 
   button #x10001 (xy- (item-size button) #x10001)
   (if enter? ccl::*gray-color* *bgc*)))

(defmethod ccl::view-mouse-enter-event-handler ((me output-stream-edit-button))
  (output-stream-edit-button-select me t))

(defmethod ccl::view-mouse-leave-event-handler ((me output-stream-edit-button))
  (output-stream-edit-button-select me nil))



(defmethod set-current-output-stream ((me output-stream-edit-button) stream
                                      &optional string)
  (declare (ignore string))
  (setf (output-stream-edit-button-stream me) stream
        (item-title me) (object-name stream))
  (ccl::invalidate-view me)
  stream)



#|
;;;
;;; test window class


(defclass testwin (ccl:window)
  ())

(defmethod ccl:view-draw-contents ((view testwin))
    (ccl:rlet ((r :rect :topleft 0 :bottomright (ccl:view-size view)))
      (ccl:with-fore-color *bgc*
        (#_fillrect r ccl:*black-pattern*))
      (call-next-method)))


;;;
;;; all controls

(make-instance 'testwin
  :view-size (xy 180 300)
  :color-p t
  :view-subviews 
  (list 
   (make-instance 'control
     :view-size (xy 40 20)
     :view-position (xy 20 20)
     :background-color *bgc*
     :action #'(lambda (me) (print me)))
   (make-instance 'button
     :view-size (xy 40 16)
     :view-position (xy 20 50)
     :background-color ccl:*red-color*
     :title "Test"
     :action #'(lambda (me) (print me)))
   (make-instance 'icon-button 
     :view-size (xy 18 18)
     :view-position (xy 20 80)
     :action #'(lambda (me) (print me)))
   (make-instance 'default-button
     :view-size (xy 46 22)
     :view-position (xy 17 110)
     :action #'(lambda (me) (print me)))
   (make-instance 'default-button
     :view-size (xy 70 22)
     :view-position (xy 17 140)
     :background-color ccl:*tan-color*
     :title "really-long-name"
     :action #'(lambda (me) (print me)))
   (make-instance 'check-box
     :view-size (xy 70 20)
     :view-position (xy 20 170)
     :background-color ccl:*orange-color*
     :title "really-long-name"
     :action #'(lambda (me) (print me)))
   (make-instance 'radio-button
     :view-size (xy 70 20)
     :view-position (xy 20 200)
     :title "Jane"
     :action #'(lambda (me) (print me)))
   (make-instance 'radio-button
     :view-size (xy 70 20)
     :view-position (xy 20 225)
     :background-color nil
     :title "Carol"
     :action #'(lambda (me) (print me)))
   (make-instance 'control
     :view-size (xy 40 20)
     :view-position (xy 100 20)
     :enabled? nil)
   (make-instance 'button
     :view-size (xy 40 16)
     :view-position (xy 100 50)
     :background-color ccl:*red-color*
     :title "really-long-name"
     :enabled? nil)
   (make-instance 'icon-button 
     :view-size (xy 18 18)
     :view-position (xy 100 80)
     :enabled? nil)
   (make-instance 'default-button
     :view-size (xy 46 22)
     :view-position (xy 97 110)
     :title "really-long-name"
     :enabled? nil)
   (make-instance 'default-button
     :view-size (xy 46 22)
     :view-position (xy 97 140)
     :background-color ccl:*purple-color*
     :title "really-long-name"
     :enabled? t
     :action #'(lambda (me) (print me)))
   (make-instance 'check-box
     :view-size (xy 70 20)
     :view-position (xy 100 170)
     :background-color ccl:*orange-color*
     :title "TesT"
     :enabled? nil)
   (make-instance 'radio-button
     :view-size (xy 70 20)
     :view-position (xy 100 200)
     :title "Kate"
     :action #'(lambda (me) (print me)))
   (make-instance 'radio-button
     :view-size (xy 70 20)
     :view-position (xy 100 225)
     :background-color ccl:*pink-color*
     :selected? t
     :title "Julia"
     :enabled? nil)
   ))
  

;;;
;;; icon buttons in various sizes

(make-instance 'testwin
  :view-size (xy 200 300)
  :color-p t
  :window-show t
  :view-subviews 
  (list (make-instance 'icon-button 
          :view-size (xy 18 18)
          :view-position (xy 20 10)
          :action #'(lambda (me) (print me)))
        (make-instance 'icon-button 
          :view-size (xy 34 34)
          :view-position (xy 20 35)
          :action #'(lambda (me) (print me)))
        (make-instance 'icon-button 
          :view-size (xy 66 66)
          :view-position (xy 20 75)
          :action #'(lambda (me) (print me)))
        (make-instance 'icon-button 
          :view-size (xy 100 100)
          :view-position (xy 20 145)
          :action #'(lambda (me) (print me)))))
             
;;;
;;; test label justifications

(make-instance 'testwin
  :view-size (xy 140 100)
  :color-p t
  :window-show t
  :view-subviews 
  (list (make-instance 'label 
          :view-size (xy 100 16)
          :view-position (xy 20 10)
          :title "Alice"
          :title-justification :left
          :framed? t)
        (make-instance 'label 
          :view-size (xy 100 16)
          :view-position (xy 20 40)
          :title "Barbara"
          :title-justification :center
          :framed? t)
        (make-instance 'label 
          :view-size (xy 100 16)
          :view-position (xy 20 70)
          :title "Claudia"
          :title-justification :right
          :framed? t)))
             


;;;
;;; different capella-fonts (update window or buttons each time)

;(set-capella-font '("Geneva" 9 :srcor :plain))
;(set-capella-font '("Times" 12 :SRCOR :italic))
;(set-capella-font '("Times" 9 :SRCOR :italic :bold :condense :underline))
;(set-capella-font '("Helvetica" 14 :srcor :bold))
;(set-capella-font '("Helvetica" 24 :srcor :bold))
      


|#

;;;
;;; -*- EOF -*-