;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

(in-package :stella)

(defconstant +layout-spacing+ 2)

;;; layout  --  the basic data structure
;;;
;;; just a name, the list of snap-on-item-stacks and the associated 
;;; snap-pane


(defvar %layouts% nil)

;; should probably be a window-owned variable to support multiple windows
(defvar *current-layout* nil)           
;(setf *current-layout* nil)           

(let ((layout-id 0))
  (defun %generate-new-layout-name (&optional suffix-str)
    (format nil "Untitled-~d~:[~;~:*~a~]" (incf layout-id) suffix-str)))

(defclass layout ()
  ((name :initform (%generate-new-layout-name)
         :initarg :name :accessor layout-name)
   (data :initform nil :initarg :data)
   (selected-snap-on-item :initform nil :initarg :selected-snap-on-item 
                  :accessor layout-selected-snap-on-item)
   (snap-on-item-container :initarg :snap-on-item-container 
                           :accessor layout-snap-on-item-container)))

(defun top-output-worksheet (&key include-invisibles)
  (find-if #'(lambda (w) 
               (typep w 'output-worksheet))
           (ccl:windows :include-invisibles include-invisibles)))

(defun set-current-layout (layout)
  ;(format t "~&; set current layout: ~a ~a" (layout-name layout) layout)
  (let ((top-output-w (top-output-worksheet))
        layout-pane snap-pane text-buffer)
    (if (and layout top-output-w)
      (progn
        (setf layout-pane (ccl:view-named ':layout-pane top-output-w) 
              text-buffer (ccl::view-named ':layout-name-text-box 
                                           layout-pane)
              snap-pane (ccl::view-named ':snap-pane layout-pane))
        (ccl::set-dialog-item-text text-buffer (layout-name layout))
        (ccl::without-interrupts
          (when (ccl::view-named ':snap-on-item-container snap-pane)
            (ccl::remove-subviews 
             snap-pane 
             (ccl::view-named ':snap-on-item-container snap-pane)))
          (ccl::add-subviews snap-pane 
                             (layout-snap-on-item-container layout)))
        (setf *current-layout* layout)
        (store/remove-layout :store))
    (setf *current-layout* layout))))

(defun get-layout-from-top-listing-view ()
  (when (and *browser* 
             (listing-browser? *browser* )
             (any-selection? *browser*))
    (set-current-layout 
     (make-layout
      :name (%generate-new-layout-name
             (format nil " (~a)"
                     (slot-value *browser* 'ccl::object-name)))
      :items (let ((objs '()))
               (map-selected-objects 
                #'(lambda (obj) 
                    (push (object-namestring obj) objs))
                (ccl::view-named ':listing-view *browser*))
               (setf objs  (list (nreverse objs))))))))



(defstruct (snap-on-item-stack (:conc-name soi-stack-))
  (layout nil :type list)               ; this one seems to be unused
  (left 0 :type fixnum)
  (width 0 :type fixnum)
  ;(default-width 0 :type fixnum)
  (items nil :type list))



;;;---------------------------------------------------------------------------
;;;
;;; snap-on-item and snap-on-item-text-item
;;;
;;; are draggable, self-positioning and self-adjusting views with an inner
;;; snap-on-item-text-item that "snap" to specific locations in the 
;;; snap-pane pane.
;;;


(defclass snap-on-item (pane)
  ((item-stack :initform nil :initarg :item-stack 
               :accessor snap-on-item-item-stack)
   (default-width :initform 0 :initarg :default-width
     :accessor snap-on-item-default-width))
  (:default-initargs 
    :background-color ccl:*white-color*))

(defclass snap-on-item-text-item (ccl::text-edit-dialog-item)
  ()
  (:default-initargs
    :text-justification 0    ; :left doesn't work
    :draw-outline nil
    :allow-returns t
    :allow-tabs t
    :view-position #x60006
    :view-font *capella-font*))


;;;
;;; movement between snap-on-items

(defvar %current-snap-on-item-row% 0)

(defun %neighbor (el list direction)
  (elt list (mod (funcall (ecase direction (:next #'1+) (:previous #'1-))
                          (position el list)) 
                 (length list))))

(defun %neighboring-stack-item (item &key (layout *current-layout*) direction)
  (let* ((stack (snap-on-item-item-stack item))
         (target-stack (%neighbor stack (slot-value layout 'data) direction))
         (target-stack-items (soi-stack-items target-stack))
         ;(my-pos (position item (soi-stack-items stack)))
         )
    (elt target-stack-items (min %current-snap-on-item-row%
                                 (1- (length target-stack-items))))))

(defun %neighboring-item (item &key (layout *current-layout*) direction)
  (let* ((stack (snap-on-item-item-stack item))
         (items (soi-stack-items stack))
         (target-pos (funcall (ecase direction (:next #'1+) (:previous #'1-)) 
                              (position item items))))
    (if (<= 0 target-pos (1- (length items)))
      (progn 
        (setf %current-snap-on-item-row% target-pos)
        (elt items target-pos))
      (let* ((target-stack (%neighbor stack (slot-value layout 'data) direction))
             (target-stack-items (soi-stack-items target-stack))
             target)
        (setf target-pos (if (plusp target-pos) 
                           0
                           (1- (length target-stack-items)))
              target (elt target-stack-items target-pos)
              %current-snap-on-item-row% target-pos)
        target))))

(defmethod ccl:view-click-event-handler :before ((self snap-on-item) where)
  (declare (ignore where))
  (setf %current-snap-on-item-row%
        (position self (soi-stack-items (snap-on-item-item-stack self)))))

(defun %set-layout-selected-snap-on-item (item)
  (setf (layout-selected-snap-on-item 
         (soi-stack-layout (snap-on-item-item-stack item))) item))

(defun %soic-forward-item (item)
  (%neighboring-item item :direction :next))

(defun %soic-backward-item (item)
  (%neighboring-item item :direction :previous))

(defun %soic-next-stack (item)
  (%neighboring-stack-item item :direction :next))

(defun %soic-previous-stack (item)
  (%neighboring-stack-item item :direction :previous))

#+(and ccl-2 (not ccl-3))
(defmacro ccl::with-text-colors (x &body body)
  (declare (ignore x))
  `(progn ,@body))
  
(defmethod ccl::view-key-event-handler ((item snap-on-item-text-item) char)
  (let ((con (ccl:view-container item))
        (win (ccl::view-window item)))
    (when (integerp char) 
      (setq char (ccl::code-char char)))
    (if (or (ccl::control-key-p)
            (member char '(#\Enter #\Newline #\Tab)))
      (let ((next-item (funcall (case (char-code char)
                                  (6 #'%soic-forward-item)      ; #\f
                                  (2 #'%soic-backward-item)     ; #\b
                                  (14 #'%soic-next-stack)       ; #\n
                                  (16 #'%soic-previous-stack)   ; #\p
                                  (9 (if (ccl::shift-key-p)     ; #\Tab
                                           #'%soic-backward-item
                                           #'%soic-forward-item))
                                  ((3 13) #'identity)   ; (#\Enter #\Newline)
                                  (t #'not))
                                con)))
        (when next-item
          (if (eq next-item con)
            (progn 
              (ccl::set-current-key-handler win 
                                            (output-worksheet-key-handler win))
              (setf (layout-selected-snap-on-item 
                     (soi-stack-layout (snap-on-item-item-stack con))) nil))
            (progn
              (ccl::set-current-key-handler win 
                                            (snap-on-item-text-item next-item))
              (%set-layout-selected-snap-on-item next-item)))))
      (ccl::with-focused-dialog-item (item con)
        (ccl::with-text-colors item
          (#_TEKey char (ccl::dialog-te-handle win)))))))


;;;
;;; appearance

(defmethod hilite-view ((me snap-on-item) &optional hilite enabled)
  (declare (ignore enabled))
  (ccl::with-font-focused-view me
    (let* ((s (item-size me))
           (w (1- (xy-x s)))
           (h (1- (xy-y s)))
           (w-2 (- w 2))
           (h-2 (- h 2)))
      (ccl:rlet ((r :rect :topleft 0 :bottomright s))
        (ccl:with-fore-color ccl:*white-color*
          (#_fillrect r ccl:*black-pattern*))
        (ccl:view-draw-contents (snap-on-item-text-item me))
        (ccl:with-fore-color (if hilite #x808080 #xb0b0b0)
          (#_insetrect r 1 1)
          (#_framerect r)))
      (ccl:with-fore-color (if hilite #xa0a0a0 #xd0d0d0)
        (#_moveto w-2 2) (#_lineto w-2 h-2) (#_lineto 2 h-2)
        (#_moveto 0 h) (#_lineto 0 0) (#_lineto (1- w) 0))
      (ccl:with-fore-color (if hilite #x202020 #x505050)
        (#_moveto 2 h-2) (#_lineto 2 2) (#_lineto (- w 3) 2)
        (#_moveto w 0) (#_lineto w h) (#_lineto 1 h)))))

(defmethod ccl::view-draw-contents ((me snap-on-item))
  (hilite-view me nil)
  ;(call-next-method)
  )


;;;
;;; mouse sensitivenenss

(defun %draw-mouse-sensitive-rect (focus-item topleft bottomright color)
  (ccl:with-focused-view focus-item
    (ccl::with-fore-color color
      (ccl:rlet ((r :rect :topleft topleft :bottomright bottomright)
                 (ps :penstate))
        (#_getpenstate ps)
        (#_pensize 2 2)
        (#_framerect r)
        (#_setpenstate ps)))))

(defun snap-on-item-select (snap-on-item enter?)
  (%draw-mouse-sensitive-rect snap-on-item #x30003
                              (xy- (item-size snap-on-item) #x30003)
                              (if enter? #xa0a0a0 ccl:*white-color*)))

(defmethod ccl::view-mouse-enter-event-handler ((me snap-on-item))
  (snap-on-item-select me t))

(defmethod ccl::view-mouse-leave-event-handler ((me snap-on-item))
  (snap-on-item-select me nil))



;;;
;;; accessors

(defmethod snap-on-item-text-item ((me snap-on-item))
  (first (ccl:subviews me)))

(defmethod snap-on-item-text ((me snap-on-item))
  (ccl::dialog-item-text (snap-on-item-text-item me)))


;;;
;;; default sizing

(defun %soi-size-from-ti (&key size text-item)
  (xy+ (or size (item-size text-item)) #xc000c))

(defun %ti-size-from-soi (&key size snap-on-item)
  (xy- (or size (item-size snap-on-item)) #xc000c))

(defmethod ccl:view-default-size ((me snap-on-item)) 
  (%soi-size-from-ti :size (ccl:view-default-size 
                            (snap-on-item-text-item me))))

(defmethod ccl:view-default-size ((me snap-on-item-text-item)) 
  (ccl:with-pstrs ((s (or (ccl:dialog-item-text me) " ")))
    (draw-using-font %capella-font%
        (let ((sw (+ 2 (#_stringwidth s)))      ; add 2 pixels--too narrow else
              (sh (+ (%font-ascent %capella-font%)
                     (%font-descent %capella-font%)
                     (%font-leading %capella-font%))))
          (xy sw sh)))))


;;;
;;; general sizing:  sizing always depends on the default text-item size,
;;; which in turn depends on the text string itself as well as on the font.

;;; cosmetic wrapper function used in ccl::exit-key-handler 
(defun size-snap-on-item (soi)
  (size-item soi nil))

(defmethod size-item ((soi snap-on-item) h &optional v)
  (let* ((ti (snap-on-item-text-item soi))
         (ds (ccl:view-default-size ti))
         (os (item-size soi))
         ns l/r) 
    ;(format t  "~& ~A ~A" (ccl:point-string ds) (ccl:point-string os))
    (if h
      (progn
        (canonicalize-point-args :point h v)
        (size-item ti (%ti-size-from-soi :size h))
        (setf ns h))
      (progn
        (size-item ti ds)
        (setf ns (%soi-size-from-ti :text-item ti))
        (setf (snap-on-item-default-width soi) (xy-x ns))))
    (setf l/r (sort (list (xy-x os) (xy-x ns)) #'<=))
    (ccl::invalidate-corners soi (xy (- (first l/r) 3) 0)
                             (xy (second l/r) (xy-y os)) t)
    (ccl:set-view-size soi ns)
    ns))

(defun size-snap-on-item-text-item (snap-on-item-text-item)
  (size-item snap-on-item-text-item 
             (ccl:view-default-size snap-on-item-text-item)))


;;;
;;; The Making Of Snap-On-Items

(defun make-snap-on-item (&rest keys)
  (let* ((ti (make-instance 'snap-on-item-text-item
               :dialog-item-text (key-arg :text keys "")))
         (tis (ccl:view-default-size ti))
         sois)
    (size-item ti tis)
    (setf sois (%soi-size-from-ti :text-item ti))
    (make-instance 'snap-on-item
      :default-width (xy-x sois)
      :view-position (key-arg :view-position keys 0)
      :view-size (key-arg :view-size keys sois)
      :view-subviews (list ti)
      :item-stack (key-arg :item-stack keys)
      :view-container (key-arg :view-container keys))))




;;;---------------------------------------------------------------------------
;;;
;;; snap-pane
;;;
;;; has an invisible subview of type snap-on-item-container which in turn
;;; hosts the current layout's snap-on-items.  this extra container is to 
;;; allow for easy swaps of the graphical layout representation (without
;;; disturbing the scrollbars and related stuff which are associated with
;;; snap-panes
;;;


;;; almost a no-op

(defclass snap-on-item-container (pane)
  ((layout :initform nil :initarg :layout 
           :accessor snap-on-item-container-layout))
  (:default-initargs
    :view-position 0
    :view-size #x7fff7fff
    :view-nick-name ':snap-on-item-container))
  
(defmethod ccl:view-draw-contents ((me snap-on-item-container))
  (declare (ignore me))
  (call-next-method))

(defclass snap-pane (pane ccl::scroller-mixin)
  ())

(defmethod ccl::view-draw-contents ((me snap-pane))
  (ccl:rlet ((r :rect :topleft 0 :bottomright #x7fff7fff))      ; (item-size me)
    (ccl:with-fore-color #x404040 ;ccl:*dark-gray-color*
      (#_fillrect r ccl:*black-pattern*)))
  (call-next-method))

(defmethod ccl::scroll-bar-page-size ((me snap-pane))
  (let ((s (item-size me)))
    (xy (round (* (xy-x s) .67)) (round (* (xy-y s) .67)))))
    
;; this one might be a bit crude...
(defmethod ccl::scroll-bar-scroll-size ((me snap-pane))
  (xy %soi-spacing% %soi-spacing%))


;;; 
;;; we don't use the scroll-pane's ccl::field-size slot, but calculate 
;;; the pane size on the fly from the current layout's space requirement.
;;; the only thing remaining to be done when adding or removing items to 
;;; the pane then is to call ccl::update-scroll-bar-limits on the snap pane
;;;   - after cut, paste and new operations
;;;   - within the exit-key-handler on snap-on-items, should a stack's width
;;;     have changed
;;;   - after a snap-on-item-snap action

(defmethod layout-size ((me layout))
  (let* ((data (slot-value me 'data))
         (last (car (last data)))
         (w (if last 
              (+ (soi-stack-left last) (soi-stack-width last) +layout-spacing+)
              0))
         ;; maximize returns 0 should data be nil
         (h (+ +layout-spacing+
               (* %soi-spacing% (loop for s in data
                                      maximize (length (soi-stack-items s)))))))
    ;(format t "~&new layout size: ~ax~a" w h)
    (xy w h)))

(defmethod layout-size ((me snap-pane))
  (let ((soic (ccl::view-named ':snap-on-item-container me)))
    (if soic 
      (layout-size (snap-on-item-container-layout soic))
      0)))

(defmethod ccl::scroll-bar-limits ((me snap-pane))
  (let ((field-s (layout-size me))
        (s (item-size me)))
    (values (xy 0 (max (ccl::scroll-bar-setting (ccl::h-scroller me)) 
                       (- (xy-x field-s) (xy-x s))))
            (xy 0 (max (ccl::scroll-bar-setting (ccl::v-scroller me)) 
                       (- (xy-y field-s) (xy-y s)))))))

(defmethod update-scroll-bar-limits ((me layout) 
                                     &optional h-scroller v-scroller)
  (declare (ignore h-scroller v-scroller))
  (ccl::update-scroll-bar-limits 
   (ccl::view-container (layout-snap-on-item-container me))))

(defmethod update-scroll-bar-limits ((me snap-pane) 
                                     &optional h-scroller v-scroller)
  (declare (ignore h-scroller v-scroller))
  (ccl::update-scroll-bar-limits me))



;;;
;;; set layout slot if made with initial snap-on-item subviews


;;;
;;; make-layout API function.  items is a list lists of strings to be used
;;; as texts for (stacked) snap-on-items.


(defun make-layout (&key (name "- untitled -") (items ()))
  (let* ((layout (make-instance 'layout :name name))
         (soic (make-instance 'snap-on-item-container
                 :layout layout))
         (data (loop for x in items
                     for s = (make-snap-on-item-stack :layout layout)
                     with l = +layout-spacing+
                     with w
                     collect 
                     (progn
                       (setf (soi-stack-left s) l
                             (soi-stack-items s)
                             (loop for i in x
                                   for top = +layout-spacing+ 
                                   then (incf top %soi-spacing%)
                                   for soi = (make-snap-on-item
                                              :view-container soic 
                                              :view-position (xy l top)
                                              :item-stack s
                                              :text i)
                                   with maxw = 0
                                   with dw
                                   collect 
                                   (progn 
                                     (setf dw (xy-x 
                                               (ccl:view-default-size soi))
                                           (slot-value soi 'default-width) dw
                                           maxw (max dw maxw))
                                     soi)
                                   finally (setf w maxw
                                                 l (+ l w +layout-spacing+)))
                             (soi-stack-width s) w)
                       (loop for i in (soi-stack-items s)
                             do (size-item i (xy w %soi-height%)))
                       s))))
    (setf (slot-value layout 'data) data
          (layout-snap-on-item-container layout) soic)
    layout))



;; stolen from scrollers.  we need different positions and sizes

(defmethod ccl::reposition-scroll-bars ((self snap-pane) 
                                        h-scroller v-scroller 
                                        &key length position)
  (let* ((pos (item-position self))
         (size (item-size self))
         (width (xy-x size))
         (height (xy-y size)))
    (when (and pos size)                ; auto-sizing may not have happenned yet 
      (ccl:without-interrupts
       (when h-scroller
         (when position
           (position-item h-scroller (xy+ pos (xy 0 height))))
         (when length
           (ccl:set-scroll-bar-length h-scroller (1+ width))))
       (when v-scroller
         (when position
           (position-item v-scroller (xy+ pos (xy width 0))))
         (when length
           (ccl:set-scroll-bar-length v-scroller (1+ height))))))))



;;;---------------------------------------------------------------------------
;;;
;;; these things packaged up lead right to our mouse and key actions!
;;;

;;;
;;; mouse event handling---------------------
;;;

;;;
;;; dragging

(defmethod ccl:view-click-event-handler ((self snap-on-item) where)
  (declare (ignore where))
  (let* ((con (ccl:view-container self))
         (consize (item-size con))
         (isize (item-size self))
         (iwidth (xy-x isize))
         (iheight (xy-y isize))
         (ipos (item-position self))
         (ileft (xy-x ipos))
         (itop (xy-y ipos))
         (iright (+ ileft iwidth))
         (ibottom (+ itop iheight))
         (in-rect t)
         ;; mouse vars:
         ;;   imouseglobal := initial mouse coords relative to container
         ;;   imouselocal := initial mouse coords relative to item position
         ;;   mouse := current mouse position
         ;;   itemexitv := vector by which to adjust the gray outline when 
         ;;                mouse exits the item rect (ie, the diff between
         ;;                exit point and imouse)
         ;;   fmousetraveled := final mouse travel vector
         (imouseglobal (ccl:view-mouse-position con))
         (imouselocal (xy- imouseglobal ipos))
         mouse itemexitv fmousetraveled)
    (declare (dynamic-extent con consize isize iwidth iheight ipos ileft
                             itop iright ibottom in-rect imouselocal mouse
                             imouseglobal fmousetraveled itemexitv))
    ;; can't use rlet for regions, since they are stored as handles
    (let ((irgn (ccl::make-record :region)))
      (declare (dynamic-extent irgn) (ccl::unsettable irgn))
      (unwind-protect
        (ccl:rlet ((slopr :rect :topleft 0 :bottomright consize)
                   (limr :rect :topleft imouselocal :bottomright consize))
          (#_setrectrgn irgn ileft itop iright ibottom)
          (ccl::with-focused-view con
            (if (ccl::double-click-p)
              (print "opening new layout window")
              (if (loop initially (hilite-view self t t)
                      do 
                      (when (not (ccl:mouse-down-p))
                        (return in-rect))
                      (if (ccl:view-contains-point-p 
                           self (setf mouse (ccl:view-mouse-position con)))
                        (when (not in-rect)
                          (setf in-rect t)
                          (hilite-view self t t)
                          (#_setrectrgn irgn ileft itop iright ibottom))
                        (progn
                          (when in-rect
                            (setf in-rect nil)
                            (hilite-view self nil t)
                            (setf itemexitv (xy- mouse imouseglobal))
                            (#_offsetrgn irgn 
                             (xy-x itemexitv) (xy-y itemexitv)))
                          ;; we need the final mouse position, so it's easier
                          ;; to simply call view-mouse-position once more 
                          ;; than to do a (xy+ (xy+ imouseglobal itemexitv)
                          ;;                        fmousetraveled))
                          ;; use fmousetraveled only for the out-of-bounds
                          ;; check
                          (setf fmousetraveled
                                (#_draggrayrgn irgn mouse limr slopr 0
                                 (ccl::%null-ptr))
                                mouse (ccl:view-mouse-position con))))
                      finally (return in-rect))
              (progn 
                (hilite-view self nil t)
                (when (ccl:view-contains-point-p (snap-on-item-text-item self)
                                                 (ccl:view-mouse-position self))
                  (%set-layout-selected-snap-on-item self)
                  (call-next-method)))
              (snap-on-item-snap self mouse)))))
        (ccl::dispose-record irgn)))))




;;;
;;; snapping

;;;
;;; a snap-pane pane is divided in horizontal zones and vertical
;;; subzones.  when the mouse is released, we first determine which snap-
;;; on-item-stack has to handle it, then whether the object should go
;;; before, right in, or after that stack, depending on which (horizontal)
;;; quarter of the stack's width the mouse is in: 
;;;
;;;   < .25      :=  insert item as a new stack before current stack
;;;   .25 - .75  :=  (re)position item in the current stack according to
;;;                  the mouse's vertical subzone
;;;   > .75      :=  insert item as a new stack after the current stack
;;;
;;; the mouse's vertical subzone is calculated as:
;;;
;;;   (min (length (soi-stack-items <soi-stack>)) (round <vpos> <itemheight>))
;;;


(defmacro linsert (el list pos)
  `(setf ,list
         (if (<= ,pos 0)
           (cons ,el ,list)
           (progn 
             (let* ((l (nthcdr (1- (min ,pos (length ,list))) ,list))
                    (tail (cons ,el (cdr l))))
               (setf (cdr l) tail))
             ,list))))

(defmacro lremove (el list &optional warn)
  `(progn
     (if (find ,el ,list)
       (setf ,list (delete ,el ,list))
       (when ,warn
         (warn "~a not found: ~a" ,el ,list)))
     ,list))


;;;
;;; layout operations

;;; 
;;; shift functions

;; stacks:  shift each stack by delta and reposition it's items

(defun shift-layout-stacks (stacks delta)
  (loop for s in stacks
        for l = (incf (soi-stack-left s) delta)
        do
        (loop for i in (soi-stack-items s)
              with y = +layout-spacing+
              do (position-item i (xy l y))
              (incf y %soi-spacing%))))

;; items:  shift items downwards by delta

(defun shift-soi-stack-items (items delta)
  (when items 
    (loop for i in items
          with l = (soi-stack-left (snap-on-item-item-stack (first items)))
          with y = (+ (xy-y (item-position (first items))) delta)
          do (position-item i (xy l y))
          (incf y %soi-spacing%))))

;;
;; update a stack's width: set all items to the default width of the widest 
;; item and update the width slot, if necessary.  shift trailing stacks by 
;; (- new-width old-width).

(defun update-soi-stack-width (stack &optional width)
  (let* ((items (soi-stack-items stack))
         (ow (soi-stack-width stack))
         (nw (or width (loop for i in items
                             maximize (snap-on-item-default-width i)))))
    (unless (= ow nw)
      (loop for i in items
            with s = (xy nw %soi-height%)
            do (size-item i s))
      (setf (soi-stack-width stack) nw)
      (shift-layout-stacks 
       (cdr (member stack (slot-value *current-layout* 'data))) 
       (- nw ow)))))


;;;
;;; removal

;; stacks: remove, shift trailing stacks left by removed stack's width plus 
;; +layout-spacing+.  we don't update removed stack, since stacks don't get
;; reused.  return new layout data

(defun remove-soi-stack (stack)
  (shift-layout-stacks 
   (cdr (member stack (slot-value *current-layout* 'data)))
   (- (+ +layout-spacing+ (soi-stack-width stack))))
  (lremove stack (slot-value *current-layout* 'data) t))


;; items: remove, move trailing items up to close the gap and update the 
;; stack's width if the removed item's default width is >= the stack width.
;; Update the vital information of the removed item, tho: reset the width
;; to it's default value.  return removed item

(defun remove-snap-on-item (item stack)
  (if (= (length (soi-stack-items stack)) 1)
    (remove-soi-stack stack)
    (progn
      (shift-soi-stack-items (cdr (member item (soi-stack-items stack)))
                             (- %soi-spacing%))
      (lremove item (soi-stack-items stack) t)
      (when (>= (snap-on-item-default-width item) (soi-stack-width stack))
        (update-soi-stack-width stack))
      (size-snap-on-item item)))
  item)


;;;
;;; insertion

;; stacks:  determine and set the left slot, position the first (single) item
;; at the base-line, insert it in the layout.  we do NOT update stack widths,
;; assuming stacks that are to be inserted have their width already set to
;; their (single) item's default width.
;; finally shift trailing stacks by our width plus the spacing constant

(defun insert-soi-stack (stack position)
  (let ((layout-data (slot-value *current-layout* 'data)) last)
    (setf (soi-stack-left stack) 
          (if (<= position 0)
            +layout-spacing+
            (if (<= position (1- (length layout-data)))
              (soi-stack-left (elt layout-data position))
              (progn 
                (setf last (car (last layout-data)))
                (if last 
                  (+ +layout-spacing+ (soi-stack-left last)
                     (soi-stack-width last))
                  +layout-spacing+)))))
    (position-item (first (soi-stack-items stack))
                   (soi-stack-left stack) +layout-spacing+)
    (linsert stack (slot-value *current-layout* 'data) position)
    ;(update-soi-stack-width stack)
    (shift-layout-stacks (nthcdr (1+ position) 
                                 (slot-value *current-layout* 'data)) 
                         (+ +layout-spacing+ (soi-stack-width stack)))))


;; items:  position and insert item at position and  shift trailing items 
;; downwards.  then link it to the new stack, and either size it if it is
;; smaller than the stack's width or update the whole stack with the new
;; items (default) width

(defun insert-snap-on-item (item stack position)
  (position-item item (xy (soi-stack-left stack) 
                          (+ +layout-spacing+ (* position %soi-spacing%))))
  (shift-soi-stack-items 
   (nthcdr (1+ position) (linsert item (soi-stack-items stack) position))
   %soi-spacing%)
  (setf (snap-on-item-item-stack item) stack)
  (let ((item-default-width (snap-on-item-default-width item))
        (stack-width (soi-stack-width stack)))
    (if (<= item-default-width stack-width)
      (size-item item (xy stack-width %soi-height%))
      (update-soi-stack-width stack item-default-width))))


#|
;;; print debugging information
(defun print-layout-data (data)
  (multiple-value-bind 
    (m d)
    (loop for s in data
          for i = (map 'list #'(lambda (x)
                                 (format nil "[~ax~a: (~a)]"
                                         (xy-x (item-size x)) 
                                         (xy-y (item-size x))
                                         (snap-on-item-default-width x)))
                       (soi-stack-items s))
          maximize (length i) into %m
          collect (cons (soi-stack-left s)
                        (cons (soi-stack-width s)
                              (cons "----------" i))) into %d
          finally (return (values (+ 3 %m) %d))) 
    (loop repeat m
        do (format t "~&~{~@[~a~]~0,25T~}" 
                   (loop for s on d 
                         collect (caar s)
                         do (setf (car s) (cdar s)))))
    (values)))
|#


(defun snap-on-item-snap (item pos)
  (cm::without-interrupts             ; no spurious redraws
    ;; setup before item is removed:
    (let* ((layout *current-layout*)
           (x (xy-x pos))
           (y (xy-y pos))
           (layout-data (slot-value layout 'data))
           (my-stack (snap-on-item-item-stack item))
           (my-id (position (snap-on-item-item-stack item) layout-data))
           (target-stack-id (or (position x layout-data
                                           :key #'soi-stack-left 
                                           :test #'>= 
                                           :from-end t)
                                 0))
           (target-stack (elt layout-data target-stack-id))
           ;; don't use target-stack directly: it might be the one
           ;; that's being removed.  we need these values to compute 
           ;; mouse-h-%
           (target-stack-left (soi-stack-left target-stack))
           (target-stack-width (soi-stack-width target-stack))
           (mouse-h-% (/ (- x target-stack-left) target-stack-width))
           (doit t))
      ;; there is a special case when item is the only item in its stack
      ;; and its stack is the target stack.  this is equivalent to a no-change!
      ;; also, optimize for reinserts
      ;(print my-stack) (print target-stack-id) (print my-id)
      (when (= (length (soi-stack-items my-stack)) 1)
        (if (or (= target-stack-id my-id)
                (and (= (- target-stack-id my-id) 1)
                     (< mouse-h-% .25))
                (and (= (- target-stack-id my-id) -1)
                     (> mouse-h-% .75)))
          (setf doit nil)
          (when (> target-stack-id my-id)
              ;(print "Decrementing target-stack-id")
              ;; we removed our stack, so decrement higher target-stack-id's
              (decf target-stack-id))))
      (when doit
        ;; else remove item. item's and stack's sizes are now updated 
        ;; (although not visible yet)
        (remove-snap-on-item item (snap-on-item-item-stack item))
        ;(format t "~&my-id: ~a  target-id: ~a mouse-h-%: ~,2f" 
        ;        my-id target-stack-id mouse-h-%)
        ;(print-layout-data (slot-value *current-layout* 'data))
        (if (<= .25 mouse-h-% .75)
          ;; insert item in current stack
          (insert-snap-on-item item target-stack
                               (min (length (soi-stack-items target-stack))
                                    (round y %soi-spacing%)))
          ;; else insert item as a new stack before or after current stack
          (let ((insert-stack (make-snap-on-item-stack      
                               ;; left is set by insert-soi-stack
                               ;; initial width is the item's default width
                               :width (snap-on-item-default-width item)
                               :layout layout
                               :items (list item))))
            (setf (snap-on-item-item-stack item) insert-stack)
            (insert-soi-stack insert-stack 
                              (if (< mouse-h-% .25) 
                                target-stack-id
                                (1+ target-stack-id)))))
        (update-scroll-bar-limits layout))))
  ;(print-layout-data (slot-value *current-layout* 'data))
  )



;;;
;;; key event handling-----------------------
;;;

(defmethod ccl::exit-key-handler ((me snap-on-item-text-item) next-item)
  (declare (ignore next-item)) 
  ;(call-next-method)
  (let ((soi (ccl:view-container me)))
    (when (typep soi 'snap-on-item)
      (let* ((stack (snap-on-item-item-stack soi))
             (ds (ccl:view-default-size soi))
             (dw (xy-x ds))
             (sw (soi-stack-width stack))
             (odw (snap-on-item-default-width soi)))
        (setf (snap-on-item-default-width soi) dw)
        ;(format t "~&sw: ~a odw: ~a dw: ~a" sw odw dw) 
        (if stack
          (if (> dw sw) 
            (progn 
              (update-soi-stack-width stack dw)
              (update-scroll-bar-limits (soi-stack-layout stack)))
            ;; dw <= sw 
            (if (= odw sw)
              (progn 
                (update-soi-stack-width stack)
                (update-scroll-bar-limits (soi-stack-layout stack)))
              (let ((ns (xy sw %soi-height%)))
                (unless (xy= (item-size soi) ns)
                  (size-item soi ns)))))
          (size-snap-on-item soi))
        ;(format t "~&~a ~a" (ccl:point-string (item-position me))
        ;        (ccl:point-string (item-size me)))
        )
      (let ((o (read-from-string (ccl::dialog-item-text me) nil nil)))
        (unless (find-object o)
          (format t "~&;Warning: Object ~s: not found." o)
          ;;  :( warn doesn't print anything and format on *error-output* 
          ;;     fires up a new listener in #+ccl3
          ;;(warn "Object ~s: not found." o)
          ))))
  t)




;;;---------------------------------------------------------------------------
;;;
;;; the panes
;;;
;;;---------------------------------------------------------------------------


;;;---------------------------------------------------------------------------
;;;  
;;; titled-3d-frame-view
;;;

;;; superclass of layout-pane and output-pane

(defclass titled-3d-frame-view (pane gray-3d-framed-view-mixin)
  ()
  (:default-initargs 
    :background-color *bgc*))


(defun %position-titled-3d-frame-view-title (item view)
  (position-item item (xy (round (/ (- (xy-x (item-size view)) 
                                       (xy-x (item-size item)))
                                    2))
                          4)))

(defun make-titled-3d-frame-view (class &rest keys)
  (let* ((label-size-correction (xy 6 0))
         v vs l ls)
    (setf l (make-instance 'label
              :title (key-arg :title keys "")
              :title-justification :center
              :view-nick-name ':title
              :background-color *bgc*)
          ls (xy+ (ccl:view-default-size l) label-size-correction)
          v (make-instance class
              :view-nick-name (key-arg :view-nick-name keys nil)
              :view-subviews (cons l (key-arg :view-subviews keys nil)))
          vs (key-arg :view-size keys (ccl:view-default-size v)))
    (setf (slot-value v 'ccl:view-size) vs)          ; (size-item v vs)
    (position-item v (key-arg :view-position keys 0))
    (setf (slot-value l 'ccl:view-size) ls)     ; (size-item l ls)
    (%position-titled-3d-frame-view-title l v)
    (setf (slot-value v 'borders) (list 4 (+ 4 (round (/ (xy-y ls) 2))) 4 4))
    v))

(defmethod frame-3d ((me titled-3d-frame-view) borders width height 
                     &optional erase?)
  (let ((left (first borders))
        (top (second borders))
        (bottom (- height (fourth borders)))
        (right (- width (third borders))))
    (%rect-frame-3d left top right bottom :lower erase?)
    (%rect-frame-3d (1+ left) (1+ top) (1- right) (1- bottom) :raise erase?)
    ;; don't forget to draw the title
    (ccl:view-draw-contents (ccl:view-named ':title me))))

(defmethod resize-item ((me titled-3d-frame-view) h &optional v)
  (size-item me h v)
  (let ((title (ccl::view-named ':title me)))
    (%position-titled-3d-frame-view-title title me)
    (ccl::invalidate-view title)))



;;;---------------------------------------------------------------------------
;;;
;;; layout pane
;;;
;;; contains a layout library popup, name typein, store button and a
;;; snap-pane (together with it's scrollbars)
;;;

(defclass layout-pane (item ccl::view)
  ((layouts :initarg :layouts :initform nil 
            :accessor layout-pane-layouts))
  (:default-initargs
    :background-color *bgc*))

(defmethod ccl::view-draw-contents ((me layout-pane))
  (ccl:rlet ((r :rect :topleft 0 :bottomright #x7fff7fff))
    (ccl:with-fore-color (item-background-color me)
      (#_fillrect r ccl:*black-pattern*)))
  (call-next-method))

(defclass layout-pop-up (ccl::pop-up-menu)
  ()
  (:default-initargs
    :auto-update-default nil))

(defclass text-box (item ccl::text-edit-dialog-item)
  ()
  (:default-initargs 
    :background-color ccl:*white-color*
    :allow-returns t
    :view-font *capella-font*))

(defmethod item-title ((self text-box))
  (ccl:dialog-item-text self))

(defclass layout-name-text-box (text-box)
  ())

(defmethod ccl::view-key-event-handler ((self layout-name-text-box) char)
  (let ((win (ccl::view-window self)))
    (when (integerp char) (setf char (char-code char)))
    (case char 
      ((#\Enter #\Newline) 
       (ccl::set-current-key-handler win (output-worksheet-key-handler win)))
      (t (ccl::with-focused-dialog-item (self (ccl::view-container self))
           (ccl::with-text-colors self
             (#_TEKey char (ccl::dialog-te-handle win))))))))


(defmethod ccl::exit-key-handler ((me layout-name-text-box) next-item)
  (declare (ignore next-item))
  (set-layout-name *current-layout*
                   (item-title me) 
                   (ccl::find-named-sibling me ':layout-pop-up))
  t)

(defun layout-name-text-box-select (item enter?)
  (let ((pos (item-position item)))
    (%draw-mouse-sensitive-rect (ccl::view-container item)
                                (xy- pos #x20002)
                                (xy+ pos (xy+ (item-size item) #x20002))
                                (if enter? #xa0a0a0 ccl:*white-color*))))


(defmethod ccl::view-mouse-enter-event-handler ((me layout-name-text-box))
  (layout-name-text-box-select me t))

(defmethod ccl::view-mouse-leave-event-handler ((me layout-name-text-box))
  (layout-name-text-box-select me nil))


(defmethod ccl:view-draw-contents :before ((me text-box))
  (let ((pos (item-position me)))
    (ccl:rlet ((r :rect :topleft pos :bottomright (xy+ (item-size me) pos)))
      (ccl:with-fore-color (item-background-color me)
        (#_fillrect r ccl:*black-pattern*))
      ;(ccl:with-fore-color ccl:*black-color*
      ;  (#_framerect r))
      )))


(defun make-layout-menu-item (layout)
  (make-instance 'ccl:menu-item
    :menu-item-title (layout-name layout)
    :menu-item-action #'(lambda ()
                          (set-current-layout layout))))

;;;
;;; get current layout, update name, append it after %layouts% and insert it
;;; in the pop-up

(defun set-layout-name (layout name layout-pop-up) 
  (when layout
    (let ((item (ccl::find-menu-item layout-pop-up (layout-name layout))))
      (when item
        (ccl::set-menu-item-title item name))
      (setf (layout-name layout) name))))

(defun store/remove-layout (action)
  (let* ((ow (top-output-worksheet :include-invisibles t))
         (pop-up (ccl::view-named ':layout-pop-up
                                  (ccl::view-named ':layout-pane ow)))
         (text-box (ccl::find-named-sibling pop-up ':layout-name-text-box))
         (items (ccl::menu-items pop-up))
         (tail (nthcdr (position "-" items :test #'string= 
                                 :key #'ccl::menu-item-title :from-end t)
                       items)))
    (ecase action 
      (:store
       ;; does the layout structure already exist?  If so, just rewrite it
       ;; and update the menu item title
       ;(format t "~&; storing ~a ~a" 
       ;        (layout-name *current-layout*) *current-layout*)
       (if (find *current-layout* %layouts%)
         (set-layout-name *current-layout* (item-title text-box) pop-up)
         (progn
           (when *current-layout*
             (setf (layout-name *current-layout*) (item-title text-box)))
           (apply #'ccl::remove-menu-items (cons pop-up tail))
           (apply #'ccl::add-menu-items 
                  (cons pop-up 
                        (cons (make-layout-menu-item *current-layout*)
                              tail)))
           ;(format t "~&;   %layouts% before: ~a" %layouts%)
           (if (consp %layouts%)
             (nconc %layouts% (list *current-layout*))
             (setf %layouts% (list *current-layout*)))
           ;(format t "~&;   %layouts% after:  ~a" %layouts%)
           )))
      (:remove
       ;(format t "~&; removing ~a ~a" 
       ;        (layout-name *current-layout*) *current-layout*)
       (ccl::remove-menu-items pop-up 
                               (find (layout-name *current-layout*) 
                                     (ccl::menu-items pop-up)
                                     :key #'ccl::menu-item-title 
                                     :test #'string-equal))
       ;(format t "~&;   %layouts% before: ~a" %layouts%)
       (setf %layouts%
             (delete *current-layout* %layouts%))
       ;(format t "~&;   %layouts% after:  ~a" %layouts%)
       (set-current-layout (or (car (last %layouts%))
                               (make-new-layout)))))))

; (setf %layouts% nil *current-layout* nil)

;;; a layout may be generated by passing a 2-dimensional list to the
;;; :items option like in :items '(("foo") ("cell repeat 4" "foo at 6.25"
;;; "i4 @ 3.75" "opus2") ("piano-solo repeat 2 pause .75"))

(defun make-new-layout ()
  (make-layout
   :name (%generate-new-layout-name)
   :items nil))

(defun duplicate-layout (layout)
  (let* ((new-layout (make-instance 'layout 
                       :name (%generate-new-layout-name)
                       ;(concatenate 'string (layout-name layout) " copy")
                       ))
         (soic (make-instance 'snap-on-item-container
                 :layout new-layout))
         (data (loop for s in (slot-value layout 'data) 
                     for ns = (make-snap-on-item-stack
                               :layout new-layout
                               :left (soi-stack-left s)
                               :width (soi-stack-width s))
                     do (setf (soi-stack-items ns)
                              (loop for i in (soi-stack-items s) collect
                                    (make-snap-on-item
                                     :view-position (item-position i)
                                     :view-size (item-size i)
                                     :item-stack ns
                                     :view-container soic
                                     :text (snap-on-item-text i))))
                     collect ns)))
    (setf (slot-value new-layout 'data) data
          (layout-snap-on-item-container new-layout) soic)
    new-layout))


(defun make-layout-menu ()
  (let ((items (list (make-instance 'ccl:menu-item
                       :menu-item-title "-"
                       :disabled t)
                     (make-instance 'ccl:menu-item
                       :menu-item-title "Duplicate"
                       :menu-item-action 
                       #'(lambda () 
                           (set-current-layout
                            (duplicate-layout *current-layout*))))
                     (make-instance 'ccl:menu-item
                       :menu-item-title "New"
                       :menu-item-action 
                       #'(lambda () 
                           (set-current-layout
                            (make-new-layout))))
                     (make-instance 'ccl:menu-item
                       :menu-item-title "Remove"
                       :menu-item-action 
                       #'(lambda () 
                           (store/remove-layout :remove))))))

    (loop for i in %layouts%
          do (push (make-layout-menu-item i) items))
    (push (make-instance 'ccl:menu-item
            :menu-item-title "-"
            :disabled t)
          items)
    (push (make-instance 'ccl:menu-item
            :menu-item-title "Layouts"
            :disabled t)
          items)
    (make-instance 'layout-pop-up
      :menu-items items
      :item-display "Layouts"
      :view-font *capella-font*
      :view-nick-name ':layout-pop-up)))


;;; we still count frame-borders to align items with the output
;;; pane.  this used to be flexible.
(defconstant +frame-border+ (+ 4 2))

(defun make-layout-pane (width top)
  (let* ((pop-up (make-layout-menu))
         (pop-up-ds (ccl:view-default-size pop-up))
         (pop-up-x (xy-x pop-up-ds))
         (pop-up-y (xy-y pop-up-ds))
         (pane (make-instance 'layout-pane
                :view-nick-name ':layout-pane
                :view-size (xy width 200)       ; gets overwritten below
                :view-position (xy 0 top)))
         (pop-up-tl (xy (* 2 +frame-border+) +frame-border+))
         (buffer (make-instance 'layout-name-text-box
                   :view-nick-name ':layout-name-text-box))
         (buffer-ds (ccl:view-default-size (progn buffer buffer)))
         (buffer-y (xy-y buffer-ds))
         (max-item-y (max pop-up-y buffer-y))
         (snap-pane-top-offs (+ +frame-border+ max-item-y +frame-border+))
         ;; removed snap-on-items arg, so this is a real default
         ;(snap-pane-default-y (+ (* (length snap-on-items) %soi-spacing%) 2))
         ;(snap-pane-real-y (min 500 (max 200 snap-pane-default-y)))
         (snap-pane-real-y 250)
         (snap-pane (make-instance 'snap-pane
                      :view-nick-name ':snap-pane
                      :track-thumb-p t
                      :draw-scroller-outline nil
                      :view-position (xy 0 snap-pane-top-offs)
                      :view-size (xy (- width 16) 
                                     snap-pane-real-y))))
    (setf (ccl::pop-up-menu-rect pop-up) 
          (ccl:make-record :rect :topleft pop-up-tl
                           :bottomright (xy+ pop-up-tl pop-up-ds))
          (slot-value pop-up 'ccl:view-size) pop-up-ds
          (slot-value pop-up 'ccl:view-position) pop-up-tl)
    (setf (slot-value pane 'ccl:view-size)
          (xy width (+ snap-pane-top-offs snap-pane-real-y 16)))
    ;; text-boxes have a 3 pix margin, so the size has to be 6 pix narrower
    ;; and we'll add 3 pix to the left position.  vertical size is fine.
    (setf (slot-value buffer 'ccl:view-size) 
          (xy (- width (* 5 +frame-border+) pop-up-x 6)
              buffer-y)
          (slot-value buffer 'ccl:view-position)
          (xy (+ (* 3 +frame-border+) pop-up-x 3) 
              (+ +frame-border+ 
                 (round (/ (- max-item-y buffer-y) 2)))))
    (ccl::add-subviews pane pop-up buffer snap-pane
                       (ccl::h-scroller snap-pane) 
                       (ccl::v-scroller snap-pane))
    pane))


;(make-layout-pane 300 50)

;;; resize policy:
;;;   text box:   resize (but keep a minimum of 20)
;;;   s/r-button: reposition horizontally
;;;   snap-pane:  resize
;;;   scrollbars: resize and reposition

;;; 
;;; the additional height for the action button is hardcoded here
;;; as: button-h (we use the same font as for the store button) + 6 (it's
;;; a default button--ugly...) + 4 (bottom margin)

(defun size-layout-pane (layout-pane size)
  (let* ((w (xy-x size))
         (h (xy-y size))
         (lp-top-offs (xy-y (item-position layout-pane)))
         (tb (ccl:view-named ':layout-name-text-box layout-pane))
         (tb-w (max 20 (- w (* 5 +frame-border+)
                          (xy-x (item-size 
                                 (ccl:view-named ':layout-pop-up 
                                                 layout-pane))))))
         (sp (ccl:view-named ':snap-pane layout-pane))
         (sp-p (item-position sp))
         (sp-l (xy-x sp-p))
         (sp-t (xy-y sp-p))
         (sp-w (max 60 (- w 15)))
         (sp-h (max (+ +layout-spacing+ %soi-spacing%)
                    (- h lp-top-offs sp-t 15)))
         (hscr (ccl::h-scroller sp))
         (vscr (ccl::v-scroller sp)))
    (size-item layout-pane w (- h lp-top-offs))
    (size-item tb (- tb-w 6) (xy-y (item-size tb)))
    (ccl::invalidate-view tb)
    (size-item sp sp-w sp-h)
    (size-item hscr (+ 2 sp-w) 16)
    (position-item hscr (1- sp-l) (+ sp-t sp-h)) 
    (size-item vscr 16 (1+ sp-h))
    (position-item vscr (+ sp-l sp-w) sp-t)
    (ccl::update-scroll-bar-limits sp)
    ))

;;;---------------------------------------------------------------------------
;;; 
;;; output pane
;;;
;;; holds a pop-up menu that's an output stream library followed by a 
;;; hierarchical syntax menu which offers possible selections for new 
;;; streams, plus a button displaying the currently selected stream.

;;;
;;; layout view contains snap-pane, a popup, typein and button

(defclass output-pane (titled-3d-frame-view)
  ())

(defclass output-stream-pop-up (ccl::pop-up-menu)
  ()
  (:default-initargs
    :auto-update-default nil))

(defmethod current-output-stream ((me output-pane))
  (let ((b (ccl::view-named ':output-stream-button me)))
    (and b (output-stream-edit-button-stream b))))

(defmethod set-current-output-stream ((me output-pane) stream
                                      &optional string)
  (declare (ignore string))
  (let ((button (ccl:view-named ':output-stream-button me)))
    (and button 
         (set-current-output-stream button stream )
         stream)))

(defun make-output-stream-menu-item (stream &optional title)
  (setf title (object-name stream))
  (make-instance 'object-menu-item
    :menu-item-title title
    :object stream 
    :menu-item-action
    #'(lambda ()
        (set-current-output-stream
         (first (output-worksheets)) stream ))))

(defun install-new-output-stream (stream)
  (let ((w (first (output-worksheets))))
    (set-current-output-stream
     (ccl:view-named ':output-stream-button 
                     (ccl:view-named ':output-pane w))
     stream )))

(defun make-stream-types-menu ()
  (let ((items '()))
    (map-subobjects
     #'(lambda (s)
         (dolist (e (slot-value s 'stream-types))
           (let ((c (find-class (first e))))
             (push
              (make-instance 'ccl:menu-item
                :menu-item-title
                (string-capitalize
                 (substitute #\Space #\- (symbol-name (first e))))
                :menu-item-action
                #'(lambda (&aux s)
                    (when (setf s (open-event-stream c))
                      (install-new-output-stream s))))
              items))))
     .syntaxes.)
    (nreverse items)))

(defun make-stream-menu ()
  (let ((items (list (make-instance 'ccl:menu-item 
                       :menu-item-title "-"
                       :disabled t)
                     (make-instance 'ccl:menu
                       :menu-item-title "New Stream"
                       :menu-items (make-stream-types-menu)))))
    (push (make-instance 'ccl:menu-item
            :menu-item-title "-"
            :disabled t)
          items)
    (push (make-instance 'ccl:menu-item
            :menu-item-title "Streams"
            :disabled t)
          items)
    (make-instance 'output-stream-pop-up
      :update-function 
      #'(lambda (m) (update-streams-menu m #'make-output-stream-menu-item))
      :menu-items items
      :item-display "Streams"
      :view-font *capella-font*
      :view-nick-name ':output-pop-up)))

(defun make-ouput-pane (window-x)
  (let* ((pane (make-titled-3d-frame-view 
                'output-pane
                :view-nick-name ':output-pane
                :title "Stream" 
                :view-size (xy window-x 50)   ; gets overwritten below
                :view-position 0))
         (streamb (make-instance 'output-stream-edit-button))
         (outputb (make-instance 'output-worksheet-key-handler
                    :title "Output"
                    :action #'(lambda (self)
                                (start-output 
                                 (worksheet-key-handler-worksheet self)))
                    :view-nick-name ':worksheet-output-button))
         (pop-up (make-stream-menu))
         (pop-up-ds (ccl:view-default-size pop-up))
         (pop-up-x (xy-x pop-up-ds))
         (pop-up-y (xy-y pop-up-ds))
         (title (ccl::view-named ':title pane))
         (frame-border (+ 2 (car (slot-value pane 'borders))))
         (title-bottom-plus-gap (+ (xy-y (item-position title))
                                   (xy-y (item-size title))
                                   frame-border))
         (pop-up-tl (xy (* 2 frame-border) title-bottom-plus-gap))
         (streamb-ds (ccl::view-default-size streamb))
         (outputb-ds (ccl::view-default-size outputb))
         (streamb-y (round (* 1.4 (xy-y streamb-ds))))
         (outputb-y (xy-y outputb-ds))
         (outputb-x (xy-x outputb-ds)))
    (setf (ccl::pop-up-menu-rect pop-up) 
          (ccl:make-record :rect :topleft pop-up-tl
                           :bottomright (xy+ pop-up-tl pop-up-ds))
          (slot-value pop-up 'ccl:view-size) pop-up-ds
          (slot-value pop-up 'ccl:view-position) pop-up-tl)
    (setf (slot-value pane 'ccl:view-size)
          (xy window-x 
              (+ title-bottom-plus-gap (max pop-up-y streamb-y outputb-y)
                 (* 2 frame-border))))
    (setf (slot-value streamb 'ccl:view-size)
          (xy (max 20
                   (- window-x (* 6 frame-border) pop-up-x outputb-x))
              streamb-y))
    (setf (slot-value outputb 'ccl:view-size) outputb-ds)
    (setf (slot-value streamb 'ccl:view-position)
          (xy (+ (* 3 frame-border) pop-up-x) 
              (+ title-bottom-plus-gap 
                 (round (/ (- pop-up-y streamb-y) 2)))))
    (setf (slot-value outputb 'ccl:view-position)
          (xy (- window-x (* 2 frame-border) outputb-x) 
              (+ title-bottom-plus-gap 
                 (round (/ (- pop-up-y outputb-y) 2)))))
    ;;; add outputb BEFORE streamb to avoid updating/erase weirdnesses!
    (ccl::add-subviews pane pop-up outputb streamb)
    pane))

;(make-ouput-pane 200)

;;;
;;; resizing affects the width of the output-stream-button and the position
;;; of the worksheet-output-button.  regarding the pane, it only affects its
;;; width
  
(defun size-output-pane (output-pane width)
  (let* ((s (item-size output-pane))
         (sb (ccl:view-named ':output-stream-button output-pane))
         (ob (ccl:view-named ':worksheet-output-button output-pane))
         (sbs (item-size sb))
         ;(obs (item-size ob))
         (obp (item-position ob))
         (sbx (xy-x sbs))
         (sby (xy-y sbs))
         (diff (- width (xy-x s)))
         (nsbx (+ sbx diff)) 
         (nsbs (xy nsbx sby)))
    (resize-item output-pane width (xy-y s))
    (size-item sb nsbs)
    (position-item ob (xy (+ (xy-x obp) diff) (xy-y obp)))
    (if (> diff 0)
      (ccl::invalidate-corners sb (xy (1- sbx) 0) sbs)
      (when (< diff 0)
        (ccl::invalidate-corners sb (xy (1- nsbx) 0) nsbs)))))
  


;;;---------------------------------------------------------------------------
;;;
;;; output worksheet

(defclass output-worksheet (worksheet)
  ((key-handler :initform nil :accessor output-worksheet-key-handler))
  (:default-initargs
    :window-title "Output"
    :window-show nil
    :grow-icon-p t
    ))

(defmethod initialize-instance :after ((self output-worksheet) &rest rest)
  (declare (ignore rest))
  (let ((width 250)
        (height 200) 
        o-y o l)
    (ccl:add-subviews self
                      (prog1 (setf o (make-ouput-pane width))
                        (setf o-y (xy-y (item-size o))) )
                      (setf l (make-layout-pane width o-y)))
    (size-item self (xy width (+ o-y (xy-y (item-size l)))))
    (let ((kh (ccl:view-named ':worksheet-output-button 
                              (ccl:view-named ':output-pane self)))
          ;; moved the key-handler to the output pane
          ;(kh (make-instance 'output-worksheet-key-handler :worksheet self))
          )
      (setf (output-worksheet-key-handler self) kh)
      (setf (worksheet-key-handler-worksheet kh) self)
      (ccl::add-key-handler kh self))
    (ccl:set-view-size self width height)))

(defmethod initialize-worksheet ((w output-worksheet))
  ;; uh-oh, we should probably have a flag to indicate whether the 
  ;; user clicked the layout-button in a listing view, but for now:
  ;; this method gets called twice, at button-click time and at 
  ;; activation time.  we ONLY want to get the toplevel selection
  ;; when there is already an output window (if not, set-current-layout
  ;; fails) AND it EITHER doesn't have a layout yet (initialization)
  ;; OR when a listing view is in front, ie. if it's an wxplicit
  ;; layout-button click.
  (when (and (top-output-worksheet)
             (or (typep (ccl::front-window) 'listing-browser)
                 (not (current-layout w))))
    (get-layout-from-top-listing-view))
  ;(update-worksheet w)
  )

(defmethod ccl:view-activate-event-handler :after ((view output-worksheet))
  nil)

(defmethod update-worksheet ((w output-worksheet))
  (declare (ignore w))
  ;(let ((b (ccl:view-named ':worksheet-output-button 
  ;                         (ccl:view-named ':output-pane w))))
  ;(if (window-send-output? w)
  ;  (control-enable b)
  ;  (control-disable b)))
  )


;;;
;;; these two are for Cmd-U support

(defmethod window-send-output ((w output-worksheet) &optional s o)
  (declare (ignore s o))
  (start-output w))

(defmethod window-send-output? ((w output-worksheet) &optional s o)
  (declare (ignore s o))
  (and (current-output-stream w)
       (current-layout w)
       (slot-value (current-layout w) 'data)))

(defmethod update-window-for-changed-object ((w output-worksheet) object)
  (let ((s (current-output-stream w)))
    (when (eq object s) (set-current-output-stream w s))))

;;; fuck, fuck, fuck!  mcl key handler methods expect every key handler 
;;; to be also a simple view!  why do they provide a mixin then??

(defclass worksheet-key-handler (item ccl::key-handler-mixin)
  ((worksheet :initform nil :initarg :worksheet 
              :accessor worksheet-key-handler-worksheet))
  (:default-initargs 
    :view-size 0
    :view-position (xy -32768 -32768)
    :allow-returns t
    :allow-tabs t))


(defun output-worksheets ()
  (ccl::windows :class 'output-worksheet :include-invisibles t))

(defmethod current-output-stream ((me t))
  (let ((w (first (output-worksheets))))
    (and w (current-output-stream w))))

(defmethod current-output-stream ((me output-worksheet))
  (let ((p (ccl:view-named ':output-pane me)))
    (and p (current-output-stream p))))

(defmethod set-current-output-stream ((me output-worksheet) stream
                                      &optional string)
  (declare (ignore string))
  (let ((button (ccl:view-named ':output-stream-button 
                                (ccl:view-named ':output-pane me))))
    (and button 
         (prog1 (set-current-output-stream button stream )
           ;(initialize-worksheet me)
           )
         stream)))

(defmethod current-layout ((me output-worksheet))
  (let ((p (ccl:view-named ':layout-pane me)))
    (and p (current-layout p))))

(defmethod current-layout ((me layout-pane))
  (let ((sp (ccl:view-named ':snap-pane me)))
    (and sp (current-layout sp))))
         
(defmethod current-layout ((me snap-pane))
  (let ((soic (ccl:view-named ':snap-on-item-container me)))
    (and soic (current-layout soic))))
         
(defmethod current-layout ((me snap-on-item-container))
  (snap-on-item-container-layout me))

         
;;;
;;; the action

(defun parse-snap-on-item-stack (soi-stack-items)
  (let ((need-merge? (> (length soi-stack-items) 1))
        (objs'())
        (starts '())
        merge)
    (flet ((parse-item (item)
             (let ((str (snap-on-item-text item)))
               (with-input-from-string (s str)
                 (let ((obj (find-object (read s nil ""))) 
                       (start 0) repeat pause)
                   (loop for form = (read s nil nil)
                         while form 
                         do (ecase form
                              ((@ at start) 
                               (setf start (read s t nil)
                                     need-merge? t))
                              ((repeat %) 
                               (setf repeat (read s t nil)))
                              ((pause --) 
                               (setf pause (read s t nil)))))
                   (when pause
                     (if (< pause 0) 
                       (error "Bogus negative pause ~a in ~s." pause str)
                       (when (= pause 0)
                         (setf pause nil)))       ; ignore zero pauses
                     (unless repeat
                       (error "Can't parse \"pause\" option without \"repeat\" ~
                               in ~s." str)))
                   (when (< start 0)
                     (error "Bogus negative start time ~a in ~s." start str))
                   (when (and repeat (< repeat 0))
                     (error "Bogus negative repeat count ~a in ~s." repeat str))
                   (when (and obj
                              (or (not repeat)          ; ignore repeat=0
                                  (> repeat 0)))
                     (when repeat
                       (let ((rest (when pause (object rest rhythm pause)))
                             (thread (thread nil (start start))))
                         (loop repeat (1- repeat)
                               do (progn
                                    (add-object obj thread)
                                    (and pause (add-object rest thread)))
                               finally (progn 
                                         (add-object obj thread)
                                         (setf obj thread)))))
                     (push obj objs)
                     (push start starts)))))))
      (loop for item in soi-stack-items
            do (parse-item item))
      ;(format t "~&~a ~a ~a" objs starts need-merge?)
      (if need-merge?
        (progn 
          (setf merge (make-object '(merge nil)
                                   :flags (logior +no-link+ +start-unset+))
                objs (nreverse objs)
                starts (nreverse starts))
          (map nil #'(lambda (o) 
                       (add-object o merge)
                       (setf (slot-value o 'start) (pop starts)
                             (slot-value o 'flags) 
                             (logior (slot-value o 'flags) +start-unset+)))
                       objs)
          ;(print (container-objects merge))
          merge)
        (car objs)))))
 
(defun start-output (output-worksheet) 
  (let ((str (current-output-stream output-worksheet))
	(layout (current-layout output-worksheet))
	data)
    (if layout
	(progn
	  (setf data (slot-value layout 'data))
	  (if data
	      (if str
                (let ((*command-prompting* nil))
		  (seq-object (loop for s in data
				    collect (parse-snap-on-item-stack 
					     (soi-stack-items s)))
			      str))
		(warn-user "~&;Warning: no output stream specified."))
	    (warn-user "~&;Warning: the current layout is empty.")))
      (warn-user "~&;Warning: no current layout."))))

(defclass output-worksheet-key-handler (worksheet-key-handler
                                        default-button)
    ())

(defmethod ccl::view-key-event-handler ((self output-worksheet-key-handler) char)
  (when (integerp char) 
    (setf char (char-code char)))
  (case char 
    ((#\Enter #\Newline) 
     (ccl::press-button self)
     ;(start-output (worksheet-key-handler-worksheet self))
     )
    #|(#\Tab (print (if (ccl::shift-key-p)
                    "selecting previous group of items"
                    "selecting next group of items")))|#
    ))


(defmethod ccl::window-close :after ((me output-worksheet))
  (let ((p1 (ccl::view-named 'output-pop-up (ccl:view-named ':output-pane me)))
        (p2 (ccl::view-named 'layout-pop-up (ccl:view-named ':layout-pane me))))
    (loop for p in (list p1 p2)
          while p
          for r = (ccl::pop-up-menu-rect p)
          when (and r (ccl::macptrp r))
          do (#_disposptr r))))


(defmethod ccl:window-size-parts ((window output-worksheet))
  (let* ((size (ccl:view-size window))
         (op (ccl:view-named ':output-pane window))
         (lp (ccl:view-named ':layout-pane window))
         (w (xy-x size)))
      (when (and op lp)
        ;; can't change the height of output-pane
        (size-output-pane op w)
        (size-layout-pane lp size))))


;;;
;;;  Command-N, X, C, V on output-worksheets

(defvar %layout-scrap% nil)

(defun %layout-scrap-copy-snap-on-item (item)
  (make-snap-on-item
   :text (snap-on-item-text item)))

(defmethod current-snap-on-item-container ((me output-worksheet))
  (ccl:view-named ':snap-on-item-container
                  (ccl::view-named ':snap-pane
                                   (ccl::view-named ':layout-pane me))))

(defmethod selected-snap-on-item ((me snap-on-item-container))
  (layout-selected-snap-on-item (snap-on-item-container-layout me)))

(defmethod selected-snap-on-item ((me output-worksheet))
  (selected-snap-on-item (current-snap-on-item-container me)))

(defmethod ccl::set-current-key-handler ((window ccl:window) (item snap-on-item)
                                         &optional (select-all t))
  (ccl::set-current-key-handler 
   window (snap-on-item-text-item item) select-all))

(defmethod ccl::cut ((me output-worksheet))
  (let ((sel (selected-snap-on-item me)))
    ;(format t "~&cut: Item ~s" (and sel (snap-on-item-text sel)))
    (when sel
      (let* ((stack (snap-on-item-item-stack sel))
             (layout (soi-stack-layout stack))
             (data (slot-value layout 'data))
             (items (soi-stack-items stack))
             (i-pos (position sel items))
             next-stack s-pos new-key-handler-item)
        ;; set key handler to next item, if any or else previous, if any
        ;; or else first item next stack, if any or else last item 
        ;; previous stack, if any and let this item be the selected one.
        ;; Else let mcl deal with it.
        (setf new-key-handler-item
              (or (cadr (member sel items))
                  (and (> i-pos 0) (elt items (1- i-pos)))
                  (and (setf next-stack (cadr (member stack data)))
                       (first (soi-stack-items next-stack)))
                  (and (setf s-pos (position stack data))
                       (> s-pos 0)
                       (car (last (soi-stack-items (elt data (1- s-pos))))))))
        ;(format t "~&new key-handler: ~s" 
        ;        (and new-key-handler-item 
        ;             (snap-on-item-text new-key-handler-item)))
        (if new-key-handler-item
          (progn
            (ccl::set-current-key-handler me new-key-handler-item)
            (%set-layout-selected-snap-on-item new-key-handler-item))
          (setf (layout-selected-snap-on-item layout) nil))
        ;(format t "~&selected-snap-on-item: ~s" 
        ;        (and (selected-snap-on-item me)
        ;             (snap-on-item-text (selected-snap-on-item me))))
        (ccl::remove-subviews (current-snap-on-item-container me) sel)
        (setf %layout-scrap%
              (remove-snap-on-item sel (snap-on-item-item-stack sel)))
        (update-scroll-bar-limits layout)
        %layout-scrap%))))

(defmethod ccl::copy ((me output-worksheet))
  (let ((sel (selected-snap-on-item me)))
    ;(format t "~&copy: Item ~s" (and sel (snap-on-item-text sel)))
    (when sel
      (setf %layout-scrap% (%layout-scrap-copy-snap-on-item sel)))))

;;;
;;; paste and new insert after currently selected item, or the end

(defun %new/paste-snap-on-item (me command)
  (let* ((c (ecase command (:new 0) (:paste 1)))
         (soic (current-snap-on-item-container me))
         (sel (selected-snap-on-item soic))
         (new (case c
                (0 (make-snap-on-item))
                (1 (%layout-scrap-copy-snap-on-item %layout-scrap%))))
         (stack (if sel
                  (snap-on-item-item-stack sel)
                  (or (car (last (slot-value *current-layout* 'data)))
                      (car (push (make-snap-on-item-stack 
                                  :layout *current-layout*
                                  :left +layout-spacing+
                                  :items nil)
                                 (slot-value *current-layout* 'data))))))
         (pos (if sel
                (1+ (position sel (soi-stack-items stack)))
                (length (soi-stack-items stack)))))
    (insert-snap-on-item new stack pos)
    (ccl::add-subviews soic new)
    (ccl::set-current-key-handler me new)
    (%set-layout-selected-snap-on-item new)
    (update-scroll-bar-limits (ccl::view-container soic))
    %layout-scrap%))

(defmethod ccl::paste ((me output-worksheet))
  (%new/paste-snap-on-item me :paste))

(defmethod ccl::new ((me output-worksheet))
  (%new/paste-snap-on-item me :new))


;;;
;;; make new work for output worksheets

(eval-when (:load-toplevel :execute)
  (let ((new (find "New" (ccl::menu-items ccl::*file-menu*) 
                   :key #'ccl::menu-item-title :test #'string=)))
    (ccl::set-menu-item-action-function 
     new
     #'(lambda (&rest rest)
         (typecase (first (ccl::windows))
           (output-worksheet (ccl::new (first (ccl::windows))))
           (t (apply #'ccl::fred rest)))))))



#|
(progn 
  (thread foo () 
    (doitems (n (notes c4 cs d ds e f in heap))
      (object midi-note note n rhythm 1/3 duration 1 amplitude .4)))
  (thread bar () 
    (doitems (n (notes fs3 fs4 fs5 fs6))
      (object midi-note note n rhythm .1 duration .1 amplitude .5))
    (object rest rhythm 2)
    (doitems (n (notes fs3 fs4 fs5 fs6))
      (object midi-note note n rhythm .1 duration .1 amplitude .5)))
  (thread baz () 
    (doitems (n (notes r (notes g5 gs a bf b in heap)))
      (object midi-note note n rhythm 1 duration 1 amplitude .25)))
  
  (thread c1 () 
    (loop repeat 5 do
          (object midi-note note 'c1 rhythm 1.9 duration .1 amplitude .4)))
  (thread c2 () 
    (loop repeat 5 do
          (object midi-note note 'c2 rhythm 1.7 duration .1 amplitude .4)))
  (thread c3 () 
    (loop repeat 5 do
          (object midi-note note 'c3 rhythm 1.3 duration .1 amplitude .4)))
  (thread c4 () 
    (loop repeat 5 do
          (object midi-note note 'c4 rhythm 1.1 duration .1 amplitude .4)))
  (thread c5 () 
    (loop repeat 5 do
          (object midi-note note 'c5 rhythm .7 duration .1 amplitude .4)))
  (thread c6 () 
    (loop repeat 5 do
          (object midi-note note 'c6 rhythm .5 duration .1 amplitude .4)))
  (thread c7 () 
    (loop repeat 5 do
          (object midi-note note 'c7 rhythm .2 duration .1 amplitude .4)))
 

;(set-current-layout 
;   (make-layout
;    :name "garbage"
;    :items '(("c7 repeat 4 pause .2")
;             ("c1" "c4 @ 1" "foo")
;             ("c7 % 3 -- .3")
;             ("c2" "c5 @.5" "bar")
;             ("c7 repeat 2 pause.5")
;             ("c3" "c6 @ .3" "baz")
;             ("c7"))))

  )

|#


